# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import copy
import re

import pyparsing as pp
import sqlalchemy as sa
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.schema import DDLElement
from sqlalchemy.sql import Select
from sqlalchemy.sql.elements import TextClause

__all__ = [
    "CreateSchema",
    "DropSchema",
    "RenameSchema",
    "DropSchemaContent",
    "CreateTableAsSelect",
    "InsertIntoSelect",
    "CreateTableWithSuffix",
    "CreateViewAsSelect",
    "CopySelectTo",
    "CreateAlias",
    "CopyTable",
    "RenameTable",
    "DropTable",
    "CreateDatabase",
    "DropAlias",
    "DropFunction",
    "DropProcedure",
    "DropView",
    "AddPrimaryKey",
    "AddIndex",
    "ChangeColumnNullable",
    "ChangeColumnTypes",
    "ChangeTableLogged",
    "LockTable",
    "LockSourceTable",
    "split_ddl_statement",
]

from sqlalchemy.sql.type_api import TypeEngine

from pydiverse.common.util.hashing import stable_hash
from pydiverse.pipedag import Schema

try:
    from sqlalchemy import Connection, Engine
except ImportError:
    # For compatibility with sqlalchemy < 2.0
    from sqlalchemy.engine import Engine
    from sqlalchemy.engine.base import Connection


# Postgres truncates identifiers at 63 characters
# MSSQL does not allow identifiers longer than 128 characters
MAX_LENGTH_PK = 63
MAX_LENGTH_INDEX = 63


class CreateSchema(DDLElement):
    def __init__(self, schema: Schema, if_not_exists=False):
        self.schema = schema
        self.if_not_exists = if_not_exists


class DropSchema(DDLElement):
    def __init__(self, schema: Schema, if_exists=False, cascade=False, *, engine=None):
        """
        :param engine: Used if cascade=True but the database doesn't support cascade.
        """
        self.schema = schema
        self.if_exists = if_exists
        self.cascade = cascade

        self.engine = engine


class RenameSchema(DDLElement):
    def __init__(self, from_: Schema, to: Schema, engine: Engine):
        self.from_ = from_
        self.to = to
        self.engine = engine


class DropSchemaContent(DDLElement):
    def __init__(self, schema: Schema, engine: Engine):
        self.schema = schema
        self.engine = engine


class CreateDatabase(DDLElement):
    def __init__(self, database: str, if_not_exists=False):
        self.database = database
        self.if_not_exists = if_not_exists


class DropDatabase(DDLElement):
    def __init__(self, database: str, if_exists=False, cascade=False):
        self.database = database
        self.if_exists = if_exists
        self.cascade = cascade


class InsertIntoSelect(DDLElement):
    def __init__(
        self,
        name: str,
        schema: Schema,
        query: Select | TextClause | sa.Text,
    ):
        self.name = name
        self.schema = schema
        self.query = query


class CreateTableAsSelect(DDLElement):
    def __init__(
        self,
        name: str,
        schema: Schema,
        query: Select | TextClause | sa.Text,
        *,
        unlogged: bool = False,
        suffix: str = "",
    ):
        self.name = name
        self.schema = schema
        self.query = query
        # Postgres supports creating unlogged tables. Flag should get ignored by
        # other dialects
        self.unlogged = unlogged
        # Suffix to be appended to the statement, e.g. from materialization details
        self.suffix = suffix


class CreateTableWithSuffix(DDLElement):
    def __init__(
        self,
        name: str,
        schema: Schema,
        sql_dtypes: dict[str, TypeEngine],
        nullable: list[str] | None,
        non_nullable: list[str] | None,
        suffix: str,
    ):
        """
        This is used for dialect=ibm_sa_db to create a table in a
        table space and with compression before we let Pandas fill it with data.
        """
        self.name = name
        self.schema = schema
        self.sql_dtypes = sql_dtypes
        self.nullable = nullable
        self.non_nullable = non_nullable
        self.suffix = suffix


class CreateViewAsSelect(DDLElement):
    def __init__(self, name: str, schema: Schema, query: Select | TextClause | sa.Text):
        self.name = name
        self.schema = schema
        self.query = query


class CopySelectTo(DDLElement):
    def __init__(self, target: str, format_spec: str, query: Select | TextClause | sa.Text):
        self.target = target
        self.format = format_spec
        self.query = query


class CreateAlias(DDLElement):
    def __init__(
        self,
        from_name,
        from_schema: Schema,
        to_name,
        to_schema: Schema,
        or_replace=False,
    ):
        self.from_name = from_name
        self.from_schema = from_schema
        self.to_name = to_name
        self.to_schema = to_schema
        self.or_replace = or_replace


class CopyTable(DDLElement):
    def __init__(
        self,
        from_name,
        from_schema: Schema,
        to_name,
        to_schema: Schema,
        *,
        unlogged: bool = False,
        suffix: str = "",
    ):
        self.from_name = from_name
        self.from_schema = from_schema
        self.to_name = to_name
        self.to_schema = to_schema
        self.unlogged = unlogged
        self.suffix = suffix


class RenameTable(DDLElement):
    def __init__(
        self,
        from_name,
        to_name,
        schema: Schema,
    ):
        self.from_name = from_name
        self.to_name = to_name
        self.schema = schema


class DropTable(DDLElement):
    def __init__(self, name, schema: Schema | str, if_exists=False, cascade=False):
        self.name = name
        self.schema = schema
        self.if_exists = if_exists
        self.cascade = cascade  # True: remove dependent views in postgres


class DropView(DDLElement):
    """
    Attention: For mssql, this statement must be prefixed with
               a 'USE <database>' statement.
    """

    def __init__(self, name, schema: Schema | str, if_exists=False):
        self.name = name
        self.schema = schema
        self.if_exists = if_exists


class DropAlias(DDLElement):
    """
    This is used for dialect=ibm_sa_db
    """

    def __init__(self, name, schema: Schema | str, if_exists=False, *, engine=None):
        """
        :param engine: Used if if_exists=True but the database doesn't support it.
        """
        self.name = name
        self.schema = schema
        self.if_exists = if_exists
        self.engine = engine


class DropNickname(DDLElement):
    """
    This is used for dialect=ibm_sa_db
    """

    def __init__(self, name, schema: Schema, if_exists=False):
        self.name = name
        self.schema = schema
        self.if_exists = if_exists


class DropProcedure(DDLElement):
    """
    Attention: For mssql, this statement must be prefixed with
               a 'USE <database>' statement.
    """

    def __init__(self, name, schema: Schema | str, if_exists=False):
        self.name = name
        self.schema = schema
        self.if_exists = if_exists


class DropFunction(DDLElement):
    """
    Attention: For mssql, this statement must be prefixed with
               a 'USE <database>' statement.
    """

    def __init__(self, name, schema: Schema | str, if_exists=False):
        self.name = name
        self.schema = schema
        self.if_exists = if_exists


def truncate_key(key_type, table_name, columns, max_length):
    # table_name + columns
    pk = key_type + table_name.lower() + "_" + columns

    if len(pk) > max_length:
        # try hash + full table name
        pk_hash = str(stable_hash(table_name.lower() + columns))
        pk = key_type + table_name.lower() + "_" + pk_hash
        if len(pk) > max_length:
            # hash + truncated table name
            pk = key_type + pk_hash + "_" + table_name.lower()[: max_length - len(key_type + pk_hash)]
    return pk


class AddPrimaryKey(DDLElement):
    def __init__(
        self,
        table_name: str,
        schema: Schema,
        key_columns: list[str],
        name: str | None = None,
    ):
        self.table_name = table_name
        self.schema = schema
        self.key = key_columns
        self._name = name

    @property
    def name(self) -> str:
        if self._name:
            return self._name
        columns = "_".join(c.lower() for c in self.key)
        return truncate_key("pk_", self.table_name, columns, MAX_LENGTH_PK)


class AddIndex(DDLElement):
    def __init__(
        self,
        table_name: str,
        schema: Schema,
        index_columns: list[str],
        name: str | None = None,
    ):
        self.table_name = table_name
        self.schema = schema
        self.index = index_columns
        self._name = name

    @property
    def name(self) -> str:
        if self._name:
            return self._name

        columns = "_".join(c.lower() for c in self.index)
        return truncate_key("idx_", self.table_name, columns, MAX_LENGTH_INDEX)


class AddClusteredColumnstoreIndex(DDLElement):
    def __init__(
        self,
        table_name: str,
        schema: Schema,
        name: str | None = None,
    ):
        self.table_name = table_name
        self.schema = schema
        self._name = name

    @property
    def name(self) -> str:
        if self._name:
            return self._name
        return "ccidx"


class ChangeColumnTypes(DDLElement):
    def __init__(
        self,
        table_name: str,
        schema: Schema,
        column_names: list[str],
        column_types: list[str],
        nullable: bool | list[bool] | None = None,
        cap_varchar_max: int | None = None,
    ):
        if not isinstance(nullable, list):
            nullable = [nullable for _ in column_names]
        self.table_name = table_name
        self.schema = schema
        self.column_names = column_names
        self.column_types = column_types
        self.nullable = nullable
        self.cap_varchar_max = cap_varchar_max


class ChangeColumnNullable(DDLElement):
    def __init__(
        self,
        table_name: str,
        schema: Schema,
        column_names: list[str],
        nullable: bool | list[bool],
    ):
        if isinstance(nullable, bool):
            nullable = [nullable for _ in column_names]
        self.table_name = table_name
        self.schema = schema
        self.column_names = column_names
        self.nullable = nullable


class ChangeTableLogged(DDLElement):
    """Changes a postgres table from LOGGED to UNLOGGED (or vice-versa)

    This reduces safety in case of a crash or unclean shutdown, but can significantly
    increase write performance:

    https://www.postgresql.org/docs/9.5/sql-createtable.html#SQL-CREATETABLE-UNLOGGED
    """

    def __init__(
        self,
        table_name: str,
        schema: Schema,
        logged: bool,
    ):
        self.table_name = table_name
        self.schema = schema
        self.logged = logged


class LockTable(DDLElement):
    def __init__(self, name: str, schema: Schema | str):
        """
        Lock Table in exclusive mode for writing
        """
        self.name = name
        self.schema = schema.get() if isinstance(schema, Schema) else schema


class LockSourceTable(DDLElement):
    def __init__(self, name: str, schema: Schema):
        """
        Lock Table in shared mode for reading.
        """
        self.name = name
        self.schema = schema.get() if isinstance(schema, Schema) else schema


@compiles(CreateSchema)
def visit_create_schema(create: CreateSchema, compiler, **kw):
    _ = kw
    schema = compiler.preparer.format_schema(create.schema.get())
    text = ["CREATE SCHEMA"]
    if create.if_not_exists:
        text.append("IF NOT EXISTS")
    text.append(schema)
    return " ".join(text)


@compiles(CreateSchema, "mssql")
def visit_create_schema_mssql(create: CreateSchema, compiler, **kw):
    _ = kw
    schema = compiler.preparer.format_schema(create.schema.get())
    if create.if_not_exists:
        unquoted_schema = create.schema.get()
        return f"""
        IF NOT EXISTS (
            SELECT 1 FROM sys.schemas WHERE name = N'{unquoted_schema}'
        )
        BEGIN EXEC('CREATE SCHEMA {schema}') END
        """
    else:
        return f"CREATE SCHEMA {schema}"


@compiles(CreateSchema, "ibm_db_sa")
def visit_create_schema_ibm_db_sa(create: CreateSchema, compiler, **kw):
    """For IBM DB2 we need to jump through extra hoops for if_exists=True."""
    _ = kw
    schema = compiler.preparer.format_schema(create.schema.get())
    if create.if_not_exists:
        return f"""
        BEGIN
            declare continue handler for sqlstate '42710' begin end;
            execute immediate 'CREATE SCHEMA {schema}';
        END
        """
    else:
        return f"CREATE SCHEMA {schema}"


@compiles(DropSchema)
def visit_drop_schema(drop: DropSchema, compiler, **kw):
    _ = kw
    schema = compiler.preparer.format_schema(drop.schema.get())
    text = ["DROP SCHEMA"]
    if drop.if_exists:
        text.append("IF EXISTS")
    text.append(schema)
    if drop.cascade:
        text.append("CASCADE")
    return " ".join(text)


@compiles(DropSchema, "mssql")
def visit_drop_schema_mssql(drop: DropSchema, compiler, **kw):
    _ = kw
    schema = compiler.preparer.format_schema(drop.schema.get())
    statements = []

    if drop.cascade:
        if drop.engine is None:
            raise ValueError(
                "Using DropSchema with cascade=True for mssql requires passing the engine kwarg to DropSchema."
            )

        statements.append(DropSchemaContent(drop.schema, drop.engine))

    text = ["DROP SCHEMA"]
    if drop.if_exists:
        text.append("IF EXISTS")
    text.append(schema)
    statements.append(" ".join(text))

    return join_ddl_statements(statements, compiler, **kw)


@compiles(DropSchema, "ibm_db_sa")
def visit_drop_schema_ibm_db_sa(drop: DropSchema, compiler, **kw):
    statements = []
    if drop.cascade:
        if drop.engine is None:
            raise ValueError(
                "Using DropSchema with cascade=True for ibm_db2 requires passing the engine kwarg to DropSchema."
            )
        statements.append(DropSchemaContent(drop.schema, drop.engine))

    # Compile DROP SCHEMA statement
    schema = compiler.preparer.format_schema(drop.schema.get())
    if drop.if_exists:
        # Add error handler to cache the case that the schema doesn't exist
        statements.append(
            f"""
            BEGIN
                declare continue handler for sqlstate '42704' begin end;
                execute immediate 'DROP SCHEMA {schema} RESTRICT';
            END
            """
        )
    else:
        statements.append(f"DROP SCHEMA {schema} RESTRICT")

    return join_ddl_statements(statements, compiler, **kw)


@compiles(RenameSchema)
def visit_rename_schema(rename: RenameSchema, compiler, **kw):
    _ = kw
    from_ = compiler.preparer.format_schema(rename.from_.get())
    to = compiler.preparer.format_schema(rename.to.get())
    return "ALTER SCHEMA " + from_ + " RENAME TO " + to


@compiles(RenameSchema, "mssql")
def visit_rename_schema_mssql(rename: RenameSchema, compiler, **kw):
    # MSSql doesn't support renaming schemas, but it allows you to move objects from
    # one schema to another.
    # https://stackoverflow.com/questions/17571233/how-to-change-schema-of-all-tables-views-and-stored-procedures-in-mssql
    # https://learn.microsoft.com/en-us/sql/t-sql/statements/alter-schema-transact-sql?view=sql-server-ver16
    from pydiverse.pipedag.backend.table.sql.reflection import (
        PipedagMSSqlReflection,
    )

    _ = kw
    from_ = compiler.preparer.format_schema(rename.from_.get())
    to = compiler.preparer.format_schema(rename.to.get())

    inspector = sa.inspect(rename.engine)

    # Reflect to get objects which we want to move
    names_to_move = []
    names_to_redefine = []

    table_names = inspector.get_table_names(schema=rename.from_.get())
    view_names = inspector.get_view_names(schema=rename.from_.get())
    alias_names = PipedagMSSqlReflection.get_alias_names(rename.engine, schema=rename.from_.get())
    procedure_names = PipedagMSSqlReflection.get_procedure_names(rename.engine, schema=rename.from_.get())
    function_names = PipedagMSSqlReflection.get_function_names(rename.engine, schema=rename.from_.get())

    names_to_move.extend(table_names)
    names_to_move.extend(alias_names)
    names_to_redefine.extend(view_names)
    names_to_redefine.extend(procedure_names)
    names_to_redefine.extend(function_names)

    if not (names_to_move or names_to_redefine):
        return ""

    # Produce statement
    statements = []

    statements.append(CreateSchema(rename.to, if_not_exists=True))
    for name in names_to_move:
        name = compiler.preparer.quote(name)
        statements.append(f"ALTER SCHEMA {to} TRANSFER {from_}.{name}")

    # Recreate views, procedures and functions, but replace all references to the
    # old schema name with the new schema name
    with rename.engine.connect() as conn:
        for name in names_to_redefine:
            definition = _mssql_update_definition(conn, name, rename.from_, rename.to)
            statements.append(definition)

    for view in view_names:
        statements.append(DropView(view, rename.from_))
    for procedure in procedure_names:
        statements.append(DropProcedure(procedure, rename.from_))
    for function in function_names:
        statements.append(DropFunction(function, rename.from_))

    statements.append(DropSchema(rename.from_))
    return join_ddl_statements(statements, compiler, **kw)


@compiles(DropSchemaContent)
def visit_drop_schema_content(drop: DropSchemaContent, compiler, **kw):
    _ = kw
    schema = drop.schema
    engine = drop.engine
    inspector = sa.inspect(engine)

    statements = []

    for table in inspector.get_table_names(schema=schema.get()):
        statements.append(DropTable(table, schema=schema))
    for view in inspector.get_view_names(schema=schema.get()):
        statements.append(DropView(view, schema=schema))

    return join_ddl_statements(statements, compiler, **kw)


@compiles(DropSchemaContent, "mssql")
def visit_drop_schema_content_mssql(drop: DropSchemaContent, compiler, **kw):
    from pydiverse.pipedag.backend.table.sql.reflection import (
        PipedagMSSqlReflection,
    )

    _ = kw
    schema = drop.schema
    engine = drop.engine
    inspector = sa.inspect(engine)

    statements = []

    for table in inspector.get_table_names(schema=schema.get()):
        statements.append(DropTable(table, schema=schema))
    for view in inspector.get_view_names(schema=schema.get()):
        statements.append(DropView(view, schema=schema))
    for alias in PipedagMSSqlReflection.get_alias_names(engine, schema=schema.get()):
        statements.append(DropAlias(alias, schema=drop.schema))
    for procedure in PipedagMSSqlReflection.get_procedure_names(engine, schema=schema.get()):
        statements.append(DropProcedure(procedure, schema=schema))
    for function in PipedagMSSqlReflection.get_function_names(engine, schema=schema.get()):
        statements.append(DropFunction(function, schema=schema))

    return join_ddl_statements(statements, compiler, **kw)


@compiles(DropSchemaContent, "ibm_db_sa")
def visit_drop_schema_content_ibm_db2(drop: DropSchemaContent, compiler, **kw):
    from pydiverse.pipedag.backend.table.sql.reflection import PipedagDB2Reflection

    _ = kw
    schema = drop.schema
    engine = drop.engine
    inspector = sa.inspect(engine)

    statements = []

    for table in inspector.get_table_names(schema=schema.get()):
        statements.append(DropTable(table, schema=schema))
    for view in inspector.get_view_names(schema=schema.get()):
        statements.append(DropView(view, schema=schema))
    for alias in PipedagDB2Reflection.get_alias_names(engine, schema=schema.get()):
        statements.append(DropAlias(alias, schema=schema))
    for nickname in PipedagDB2Reflection.get_nickname_names(engine, schema=schema.get()):
        statements.append(DropNickname(nickname, schema))

    return join_ddl_statements(statements, compiler, **kw)


@compiles(CreateDatabase)
def visit_create_database(create: CreateDatabase, compiler, **kw):
    _ = kw
    database = compiler.preparer.format_schema(create.database)
    text = ["CREATE DATABASE"]
    if create.if_not_exists:
        text.append("IF NOT EXISTS")
    text.append(database)
    return " ".join(text)


@compiles(DropDatabase)
def visit_drop_database(drop: DropDatabase, compiler, **kw):
    _ = kw
    schema = compiler.preparer.format_schema(drop.database)
    text = ["DROP DATABASE"]
    if drop.if_exists:
        text.append("IF EXISTS")
    text.append(schema)
    if drop.cascade:
        text.append("CASCADE")
    ret = " ".join(text)
    raise NotImplementedError(f"Disable for now for safety reasons (not yet needed): {ret}")


def _visit_fill_obj_as_select(create, compiler, _type, kw, *, cmd="CREATE ", sep=" AS", prefix="", suffix=""):
    name = compiler.preparer.quote(create.name)
    schema = compiler.preparer.format_schema(create.schema.get())
    kw["literal_binds"] = True
    select = compiler.sql_compiler.process(create.query, **kw)
    return f"{cmd}{_type} {schema}.{name}{sep}\n{prefix}{select}{suffix}"


@compiles(InsertIntoSelect)
def visit_insert_into_select(insert: InsertIntoSelect, compiler, **kw):
    return _visit_fill_obj_as_select(insert, compiler, "", kw, cmd="INSERT INTO", sep="")


@compiles(InsertIntoSelect, "mssql")
def visit_insert_into_select_mssql(insert: InsertIntoSelect, compiler, **kw):
    return _visit_fill_obj_as_select(insert, compiler, "", kw, cmd="INSERT INTO", sep=" WITH(TABLOCKX)")


@compiles(CreateTableAsSelect)
def visit_create_table_as_select(create: CreateTableAsSelect, compiler, **kw):
    return _visit_fill_obj_as_select(create, compiler, "TABLE", kw)


@compiles(CreateTableAsSelect, "postgresql")
def visit_create_table_as_select_postgresql(create: CreateTableAsSelect, compiler, **kw):
    if create.unlogged:
        return _visit_fill_obj_as_select(create, compiler, "UNLOGGED TABLE", kw)
    else:
        return _visit_fill_obj_as_select(create, compiler, "TABLE", kw)


@compiles(CreateTableAsSelect, "mssql")
def visit_create_table_as_select_mssql(create: CreateTableAsSelect, compiler, **kw):
    name = compiler.preparer.quote(create.name)
    schema = compiler.preparer.format_schema(create.schema.get())

    kw["literal_binds"] = True
    select = compiler.sql_compiler.process(create.query, **kw)

    return insert_into_in_query(select, schema, name)


@compiles(CreateTableAsSelect, "ibm_db_sa")
def visit_create_table_as_select_ibm_db_sa(create: CreateTableAsSelect, compiler, **kw):
    # Attention: for DB2, a CreateTableAsSelect must be followed by an InsertIntoSelect
    # to actually fill data
    suffix = ") DEFINITION ONLY " + create.suffix
    return _visit_fill_obj_as_select(create, compiler, "TABLE", kw, prefix="(", suffix=suffix)


@compiles(CreateTableWithSuffix, "ibm_db_sa")
def visit_create_table_with_suffix(create: CreateTableWithSuffix, compiler, **kw):
    _ = kw
    name = compiler.preparer.quote(create.name)
    schema = compiler.preparer.format_schema(create.schema.get())
    statement = (
        f"CREATE TABLE {schema}.{name} (\n"
        + ",\n".join(
            f"{compiler.preparer.quote(col_name)} {dtype.compile(compiler.dialect)}"
            for col_name, dtype in create.sql_dtypes.items()
        )
        + f"\n) {create.suffix}"
    )
    return statement


@compiles(CreateViewAsSelect)
def visit_create_view_as_select(create: CreateViewAsSelect, compiler, **kw):
    return _visit_fill_obj_as_select(create, compiler, "VIEW", kw)


@compiles(CopySelectTo)
def visit_copy_to_as_select(copy: CopySelectTo, compiler, **kw):
    kw["literal_binds"] = True
    select = compiler.sql_compiler.process(copy.query, **kw)
    tick = "'"
    return f"COPY (\n{select}\n) TO '{str(copy.target).replace(tick, '')}' WITH (FORMAT {copy.format})"


@compiles(CopySelectTo, "duckdb")
def visit_copy_to_as_select(copy: CopySelectTo, compiler, **kw):
    kw["literal_binds"] = True
    select = compiler.sql_compiler.process(copy.query, **kw)
    tick = "'"
    return f"COPY (\n{select}\n) TO '{str(copy.target).replace(tick, '')}' (FORMAT {copy.format})"


def insert_into_in_query(select_sql, schema, table):
    into = f"INTO {schema}.{table}"
    into_point = None
    # insert INTO before first FROM, WHERE, GROUP BY, WINDOW, HAVING,
    #                          ORDER BY, UNION, EXCEPT, INTERSECT
    for marker in [
        "FROM",
        "WHERE",
        r"GROUP\s*BY",
        "WINDOW",
        "HAVING",
        r"ORDER\s*BY",
        "UNION",
        "EXCEPT",
        "INTERSECT",
    ]:
        regex = re.compile(r"\b" + marker + r"\b", re.IGNORECASE)
        for match in regex.finditer(select_sql):
            match_start = match.span()[0]
            # special case handling: ignore escaped from
            prev = select_sql[0:match_start]
            # ignore marker in subqueries in select columns
            # as well as columns called from (escaped by [] or "")
            if prev.count("(") == prev.count(")") and prev.count("[") == prev.count("]") and prev.count('"') % 2 == 0:
                into_point = match_start
                break
        if into_point is not None:
            break
    return (
        select_sql[0:into_point] + into + " " + select_sql[into_point:]
        if into_point is not None
        else select_sql + " " + into
    )


@compiles(CreateAlias)
def visit_create_alias(create_alias: CreateAlias, compiler, **kw):
    from_name = compiler.preparer.quote(create_alias.from_name)
    from_schema = compiler.preparer.format_schema(create_alias.from_schema.get())
    query = sa.select("*").select_from(sa.text(f"{from_schema}.{from_name}"))
    return compiler.process(CreateViewAsSelect(create_alias.to_name, create_alias.to_schema, query), **kw)


@compiles(CreateAlias, "mssql")
def visit_create_alias(create_alias: CreateAlias, compiler, **kw):
    from_name = compiler.preparer.quote(create_alias.from_name)
    from_schema = compiler.preparer.format_schema(create_alias.from_schema.get())
    to_name = compiler.preparer.quote(create_alias.to_name)
    to_schema = compiler.preparer.format_schema(create_alias.to_schema.get())

    text = ["CREATE"]
    if create_alias.or_replace:
        text.append("OR REPLACE")
    text.append(f"SYNONYM {to_schema}.{to_name} FOR {from_schema}.{from_name}")
    return " ".join(text)


@compiles(CreateAlias, "ibm_db_sa")
def visit_create_alias(create_alias: CreateAlias, compiler, **kw):
    from_name = compiler.preparer.quote(create_alias.from_name)
    from_schema = compiler.preparer.format_schema(create_alias.from_schema.get())
    to_name = compiler.preparer.quote(create_alias.to_name)
    to_schema = compiler.preparer.format_schema(create_alias.to_schema.get())

    text = ["CREATE"]
    if create_alias.or_replace:
        text.append("OR REPLACE")
    text.append(f"ALIAS {to_schema}.{to_name} FOR TABLE {from_schema}.{from_name}")
    return " ".join(text)


@compiles(CopyTable)
def visit_copy_table(copy_table: CopyTable, compiler, **kw):
    from_name = compiler.preparer.quote(copy_table.from_name)
    from_schema = compiler.preparer.format_schema(copy_table.from_schema.get())
    query = sa.select("*").select_from(sa.text(f"{from_schema}.{from_name}"))
    create = CreateTableAsSelect(
        copy_table.to_name,
        copy_table.to_schema,
        query,
        unlogged=copy_table.unlogged,
        suffix=copy_table.suffix,
    )
    return compiler.process(create, **kw)


@compiles(RenameTable)
def visit_rename_table(rename_table: RenameTable, compiler, **kw):
    _ = kw
    from_table = compiler.preparer.quote(rename_table.from_name)
    to_table = compiler.preparer.quote(rename_table.to_name)
    schema = compiler.preparer.format_schema(rename_table.schema.get())
    return f"ALTER TABLE {schema}.{from_table} RENAME TO {to_table}"


@compiles(RenameTable, "mssql")
def visit_rename_table(rename_table: RenameTable, compiler, **kw):
    _ = kw

    schema = compiler.preparer.format_schema(rename_table.schema.get())
    from_table = compiler.preparer.quote(rename_table.from_name)
    to_table = rename_table.to_name  # no quoting is intentional

    return f"EXEC sp_rename '{schema}.{from_table}', '{to_table}'"


@compiles(RenameTable, "ibm_db_sa")
def visit_rename_table(rename_table: RenameTable, compiler, **kw):
    _ = kw
    from_table = compiler.preparer.quote(rename_table.from_name)
    to_table = compiler.preparer.quote(rename_table.to_name)
    schema = compiler.preparer.format_schema(rename_table.schema.get())
    return f"RENAME TABLE {schema}.{from_table} TO {to_table}"


@compiles(RenameTable, "snowflake")
def visit_rename_table(rename_table: RenameTable, compiler, **kw):
    _ = kw
    from_table = compiler.preparer.quote(rename_table.from_name)
    to_table = compiler.preparer.quote(rename_table.to_name)
    schema = compiler.preparer.format_schema(rename_table.schema.get())
    return f"ALTER TABLE {schema}.{from_table} RENAME TO {schema}.{to_table}"


@compiles(DropTable)
def visit_drop_table(drop: DropTable, compiler, **kw):
    return _visit_drop_anything(drop, "TABLE", compiler, **kw)


@compiles(DropTable, "mssql")
def visit_drop_table(drop: DropTable, compiler, **kw):
    drop.cascade = False  # not supported by dialect
    return _visit_drop_anything(drop, "TABLE", compiler, **kw)


@compiles(DropTable, "ibm_db_sa")
def visit_drop_table(drop: DropTable, compiler, **kw):
    drop.cascade = False  # not supported by dialect
    return _visit_drop_anything(drop, "TABLE", compiler, **kw)


@compiles(DropView)
def visit_drop_view(drop: DropView, compiler, **kw):
    return _visit_drop_anything(drop, "VIEW", compiler, **kw)


@compiles(DropAlias)
def visit_drop_alias(drop: DropAlias, compiler, **kw):
    # Not all dialects support a table ALIAS as a first class object.
    # For those that don't we just use views.
    return _visit_drop_anything(drop, "VIEW", compiler, **kw)


@compiles(DropAlias, "mssql")
def visit_drop_alias_mssql(drop: DropAlias, compiler, **kw):
    # What is called ALIAS for dialect ibm_db_sa is called SYNONYM for mssql
    return _visit_drop_anything(drop, "SYNONYM", compiler, **kw)


@compiles(DropAlias, "ibm_db_sa")
def visit_drop_alias_ibm_db_sa(drop: DropAlias, compiler, **kw):
    if drop.if_exists:
        from pydiverse.pipedag.backend.table.sql.reflection import PipedagDB2Reflection

        schema_str = drop.schema.get() if isinstance(drop.schema, Schema) else drop.schema
        if drop.name not in PipedagDB2Reflection.get_alias_names(drop.engine, schema=schema_str):
            return ""
        drop = DropAlias(drop.name, drop.schema, if_exists=False)
    return _visit_drop_anything(drop, "ALIAS", compiler, **kw)


@compiles(DropNickname, "ibm_db_sa")
def visit_drop_nickname_ibm_db_sa(drop: DropAlias, compiler, **kw):
    return _visit_drop_anything(drop, "NICKNAME", compiler, **kw)


@compiles(DropProcedure)
def visit_drop_table(drop: DropProcedure, compiler, **kw):
    return _visit_drop_anything(drop, "PROCEDURE", compiler, **kw)


@compiles(DropFunction)
def visit_drop_table(drop: DropFunction, compiler, **kw):
    return _visit_drop_anything(drop, "FUNCTION", compiler, **kw)


def _visit_drop_anything(
    drop: DropTable | DropView | DropProcedure | DropFunction | DropAlias,
    _type,
    compiler,
    **kw,
):
    _ = kw
    table = compiler.preparer.quote(drop.name)
    schema_str = drop.schema.get() if isinstance(drop.schema, Schema) else drop.schema
    schema = compiler.preparer.format_schema(schema_str)
    text = [f"DROP {_type}"]
    if drop.if_exists:
        text.append("IF EXISTS")
    text.append(f"{schema}.{table}")
    if hasattr(drop, "cascade") and drop.cascade:
        text.append("CASCADE")
    return " ".join(text)


@compiles(AddPrimaryKey)
def visit_add_primary_key(add_primary_key: AddPrimaryKey, compiler, **kw):
    _ = kw
    table = compiler.preparer.quote(add_primary_key.table_name)
    schema = compiler.preparer.format_schema(add_primary_key.schema.get())
    pk_name = compiler.preparer.quote(add_primary_key.name)
    cols = ",".join([compiler.preparer.quote(col) for col in add_primary_key.key])
    return f"ALTER TABLE {schema}.{table} ADD CONSTRAINT {pk_name} PRIMARY KEY ({cols})"


@compiles(AddPrimaryKey, "duckdb")
def visit_add_primary_key(add_primary_key: AddPrimaryKey, compiler, **kw):
    _ = kw
    table = compiler.preparer.quote(add_primary_key.table_name)
    schema = compiler.preparer.format_schema(add_primary_key.schema.get())
    pk_name = compiler.preparer.quote(add_primary_key.name)
    cols = ",".join([compiler.preparer.quote(col) for col in add_primary_key.key])
    return f"CREATE UNIQUE INDEX {pk_name} ON {schema}.{table} ({cols})"


@compiles(AddIndex)
def visit_add_index(add_index: AddIndex, compiler, **kw):
    _ = kw
    table = compiler.preparer.quote(add_index.table_name)
    schema = compiler.preparer.format_schema(add_index.schema.get())
    index_name = compiler.preparer.quote(add_index.name)
    cols = ",".join([compiler.preparer.quote(col) for col in add_index.index])
    return f"CREATE INDEX {index_name} ON {schema}.{table} ({cols})"


@compiles(AddIndex, "ibm_db_sa")
def visit_add_index(add_index: AddIndex, compiler, **kw):
    _ = kw
    table = compiler.preparer.quote(add_index.table_name)
    schema = compiler.preparer.format_schema(add_index.schema.get())
    index_name = compiler.preparer.quote(add_index.name)
    cols = ",".join([compiler.preparer.quote(col) for col in add_index.index])
    return f"CREATE INDEX {schema}.{index_name} ON {schema}.{table} ({cols})"


@compiles(AddClusteredColumnstoreIndex, "mssql")
def visit_add_clustered_columnstore_index(
    add_clustered_columnstore_index: AddClusteredColumnstoreIndex, compiler, **kw
) -> str:
    _ = kw
    table = compiler.preparer.quote(add_clustered_columnstore_index.table_name)
    schema = compiler.preparer.format_schema(add_clustered_columnstore_index.schema.get())
    index_name = compiler.preparer.quote(add_clustered_columnstore_index.name)
    return f"CREATE CLUSTERED COLUMNSTORE INDEX {index_name} ON {schema}.{table}"


@compiles(ChangeColumnTypes)
def visit_change_column_types(change: ChangeColumnTypes, compiler, **kw):
    _ = kw
    table = compiler.preparer.quote(change.table_name)
    schema = compiler.preparer.format_schema(change.schema.get())
    alter_columns = ",".join(
        [
            f"ALTER COLUMN {compiler.preparer.quote(col)} SET DATA TYPE {compiler.type_compiler.process(_type)}"
            for col, _type, nullable in zip(change.column_names, change.column_types, change.nullable)
        ]
        + [
            f"ALTER COLUMN {compiler.preparer.quote(col)} {'SET' if not nullable else 'DROP'} NOT NULL"
            for col, nullable in zip(change.column_names, change.nullable)
            if nullable is not None
        ]
    )
    return f"ALTER TABLE {schema}.{table} {alter_columns}"


@compiles(ChangeColumnTypes, "snowflake")
def visit_change_column_types(change: ChangeColumnTypes, compiler, **kw):
    _ = kw
    table = compiler.preparer.quote(change.table_name)
    schema = compiler.preparer.format_schema(change.schema.get())
    alter_columns = ",".join(
        [
            f"COLUMN {compiler.preparer.quote(col)} SET DATA TYPE {compiler.type_compiler.process(_type)}"
            for col, _type, nullable in zip(change.column_names, change.column_types, change.nullable)
        ]
        + [
            f"COLUMN {compiler.preparer.quote(col)} {'SET' if not nullable else 'DROP'} NOT NULL"
            for col, nullable in zip(change.column_names, change.nullable)
            if nullable is not None
        ]
    )
    return f"ALTER TABLE {schema}.{table} ALTER {alter_columns}"


@compiles(ChangeColumnTypes, "duckdb")
def visit_change_column_types_duckdb(change: ChangeColumnTypes, compiler, **kw):
    table = compiler.preparer.quote(change.table_name)
    schema = compiler.preparer.format_schema(change.schema.get())
    alter_columns = [
        f"ALTER COLUMN {compiler.preparer.quote(col)} SET DATA TYPE {compiler.type_compiler.process(_type)}"
        for col, _type, nullable in zip(change.column_names, change.column_types, change.nullable)
    ] + [
        f"ALTER COLUMN {compiler.preparer.quote(col)} {'SET' if not nullable else 'DROP'} NOT NULL"
        for col, nullable in zip(change.column_names, change.nullable)
        if nullable is not None
    ]
    statements = [f"ALTER TABLE {schema}.{table} {statement}" for statement in alter_columns]
    return join_ddl_statements(statements, compiler, **kw)


@compiles(ChangeColumnTypes, "mssql")
def visit_change_column_types(change: ChangeColumnTypes, compiler, **kw):
    _ = kw

    table = compiler.preparer.quote(change.table_name)
    schema = compiler.preparer.format_schema(change.schema.get())

    def modify_type(_type):
        if change.cap_varchar_max is not None:
            _type = copy.copy(_type)
            if isinstance(_type, sa.String) and (_type.length is None or _type.length > change.cap_varchar_max):
                # impose some limit to allow use in primary key / index
                _type.length = change.cap_varchar_max
        return _type

    statements = [
        f"ALTER TABLE {schema}.{table} ALTER COLUMN"
        f" {compiler.preparer.quote(col)} "
        f"{compiler.type_compiler.process(modify_type(_type))}"
        f"{'' if nullable is None else ' NULL' if nullable else ' NOT NULL'}"
        for col, _type, nullable in zip(change.column_names, change.column_types, change.nullable)
    ]
    return join_ddl_statements(statements, compiler, **kw)


@compiles(ChangeColumnTypes, "ibm_db_sa")
def visit_change_column_types(change: ChangeColumnTypes, compiler, **kw):
    _ = kw

    def modify_type(_type):
        if change.cap_varchar_max is not None:
            _type = copy.copy(_type)
            if isinstance(_type, sa.String) and (_type.length is None or _type.length > change.cap_varchar_max):
                # impose some limit to allow use in primary key / index
                _type.length = change.cap_varchar_max
        return _type

    table = compiler.preparer.quote(change.table_name)
    schema = compiler.preparer.format_schema(change.schema.get())
    statements = [
        f"ALTER TABLE {schema}.{table} ALTER COLUMN"
        f" {compiler.preparer.quote(col)} SET DATA TYPE"
        f" {compiler.type_compiler.process(modify_type(_type))}"
        for col, _type, nullable in zip(change.column_names, change.column_types, change.nullable)
    ] + [
        f"ALTER TABLE {schema}.{table} ALTER COLUMN"
        f" {compiler.preparer.quote(col)}"
        f" {'SET' if not nullable else 'DROP'} NOT NULL"
        for col, nullable in zip(change.column_names, change.nullable)
        if nullable is not None
    ]
    statements.append(f"call sysproc.admin_cmd('REORG TABLE {schema}.{table}')")
    return join_ddl_statements(statements, compiler, **kw)


@compiles(ChangeColumnNullable)
def visit_change_column_nullable(change: ChangeColumnNullable, compiler, **kw):
    _ = kw
    table = compiler.preparer.quote(change.table_name)
    schema = compiler.preparer.format_schema(change.schema.get())
    alter_columns = ",".join(
        [
            f"ALTER COLUMN {compiler.preparer.quote(col)} {'SET' if not nullable else 'DROP'} NOT NULL"
            for col, nullable in zip(change.column_names, change.nullable)
        ]
    )
    return f"ALTER TABLE {schema}.{table} {alter_columns}"


@compiles(ChangeColumnNullable, "snowflake")
def visit_change_column_nullable(change: ChangeColumnNullable, compiler, **kw):
    _ = kw
    table = compiler.preparer.quote(change.table_name)
    schema = compiler.preparer.format_schema(change.schema.get())
    alter_columns = ",".join(
        [
            f"COLUMN {compiler.preparer.quote(col)} {'SET' if not nullable else 'DROP'} NOT NULL"
            for col, nullable in zip(change.column_names, change.nullable)
        ]
    )
    return f"ALTER TABLE {schema}.{table} ALTER {alter_columns}"


@compiles(ChangeColumnNullable, "ibm_db_sa")
def visit_change_column_nullable(change: ChangeColumnNullable, compiler, **kw):
    _ = kw
    statements = _get_nullable_change_statements(change, compiler)
    table = compiler.preparer.quote(change.table_name)
    schema = compiler.preparer.format_schema(change.schema.get())
    statements.append(f"call sysproc.admin_cmd('REORG TABLE {schema}.{table}')")
    return join_ddl_statements(statements, compiler, **kw)


@compiles(ChangeColumnNullable, "duckdb")
def visit_change_column_nullable(change: ChangeColumnNullable, compiler, **kw):
    _ = kw
    statements = _get_nullable_change_statements(change, compiler)
    return join_ddl_statements(statements, compiler, **kw)


@compiles(ChangeTableLogged, "postgresql")
def visit_change_table_logged(change: ChangeTableLogged, compiler, **kw):
    _ = kw
    table_name = compiler.preparer.quote(change.table_name)
    schema = compiler.preparer.format_schema(change.schema.get())
    logged = "LOGGED" if change.logged else "UNLOGGED"
    return f"ALTER TABLE {schema}.{table_name} SET {logged}"


@compiles(LockTable)
def visit_lock_table_postgres(lock_table: LockTable, compiler, **kw):
    _ = kw
    preparer = compiler.preparer
    name = preparer.quote(lock_table.name)
    schema = preparer.format_schema(lock_table.schema)

    return f"LOCK TABLE {schema}.{name} IN ACCESS EXCLUSIVE MODE"


@compiles(LockTable, "ibm_db_sa")
def visit_lock_table_ibm_db_sa(lock_table: LockTable, compiler, **kw):
    _ = kw
    preparer = compiler.preparer
    name = preparer.quote(lock_table.name)
    schema = preparer.format_schema(lock_table.schema)

    return f"LOCK TABLE {schema}.{name} IN EXCLUSIVE MODE"


@compiles(LockSourceTable)
def visit_lock_source_table_postgres(lock_source_table: LockSourceTable, compiler, **kw):
    _ = kw
    preparer = compiler.preparer
    name = preparer.quote(lock_source_table.name)
    schema = preparer.format_schema(lock_source_table.schema)

    return f"LOCK TABLE {schema}.{name} IN SHARE MODE"


@compiles(LockSourceTable, "ibm_db_sa")
def visit_lock_source_table_ibm_db_sa(lock_source_table: LockSourceTable, compiler, **kw):
    _ = kw
    preparer = compiler.preparer
    name = preparer.quote(lock_source_table.name)
    schema = preparer.format_schema(lock_source_table.schema)

    return f"LOCK TABLE {schema}.{name} IN SHARE MODE"


try:
    from sqlalchemy.sql.elements import TryCast

    @compiles(TryCast, "postgresql")
    def visit_try_cast_postgresql(element, compiler, **kw):
        """
        Compile TryCast for PostgreSQL using pg_input_is_valid function.

        Uses CASE WHEN pg_input_is_valid(value, type) THEN CAST(value AS type) ELSE NULL END
        to safely attempt casting and return NULL if the cast would fail.
        """
        value = compiler.process(element.clause, **kw)
        target_type = str(element.type)

        return (
            f"CASE WHEN pg_input_is_valid({value}::VARCHAR, '{target_type}') "
            f"THEN CAST({value} AS {target_type}) ELSE NULL END"
        )
except ImportError:
    # no TryCast implementation for SQLAlchemy < 1.0
    pass


def _get_nullable_change_statements(change, compiler):
    table = compiler.preparer.quote(change.table_name)
    schema = compiler.preparer.format_schema(change.schema.get())
    statements = [
        f"ALTER TABLE {schema}.{table} ALTER COLUMN"
        f" {compiler.preparer.quote(col)}"
        f" {'SET' if not nullable else 'DROP'} NOT NULL"
        for col, nullable in zip(change.column_names, change.nullable)
    ]
    return statements


def _mssql_update_definition(
    conn: Connection,
    name: str,
    old_schema: Schema,
    new_schema: Schema,
) -> str:
    """
    Loads the definition of object `name` in `old_schema`, and replaces all references
    to `old_schema` inside the definition with `new_schema`.

    :return: updated definition referencing `new_schema` instead of `old_schema`.
    """

    # Get definition from database (using unquoted name)
    query = f"SELECT OBJECT_DEFINITION(OBJECT_ID(N'{old_schema.get()}.{name}'))"
    definition = conn.exec_driver_sql(query).scalar_one()

    # Replace unquoted names with quoted ones
    identifier_preparer = conn.dialect.identifier_preparer
    old_schema = identifier_preparer.format_schema(old_schema.get())
    new_schema = identifier_preparer.format_schema(new_schema.get())

    # Replace schema in definition with new destination schema
    string_literal = pp.QuotedString(quote_char="'", esc_quote="''")
    quoted_identifier = pp.QuotedString(quote_char="[", esc_quote="]]", end_quote_char="]")

    schema_expr = pp.CaselessKeyword(old_schema).ignore(string_literal)
    if old_schema.startswith("["):
        schema_expr = schema_expr.ignore(quoted_identifier)
    else:
        schema_expr = schema_expr.ignore(quoted_identifier) | (pp.Literal("[") + schema_expr + pp.Literal("]"))

    schema_expr = schema_expr.set_parse_action(pp.replace_with(new_schema))
    expr = schema_expr + pp.FollowedBy(".")

    return expr.transform_string(definition)


STATEMENT_SEPARATOR = "; -- PYDIVERSE-PIPEDAG-SPLIT\n"


def join_ddl_statements(statements, compiler, **kw):
    """Mechanism to combine multiple DDL statements into one."""
    statement_strings = []
    kw["literal_binds"] = True
    for statement in statements:
        if isinstance(statement, str):
            statement_strings.append(statement)
        else:
            statement_strings.append(compiler.process(statement, **kw))
    return STATEMENT_SEPARATOR.join(statement_strings)


def split_ddl_statement(statement: str):
    """Split previously combined DDL statements apart"""
    return [statement for statement in statement.split(STATEMENT_SEPARATOR) if statement.strip() != ""]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.pipedag.materialize.debug import materialize_table

__all__ = ["materialize_table"]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import copy
import inspect
import typing
from collections.abc import Iterable
from functools import total_ordering
from typing import TYPE_CHECKING, Any, Generic, overload

import sqlalchemy as sa
import structlog
from attr import frozen

from pydiverse.pipedag._typing import T
from pydiverse.pipedag.context import ConfigContext, TaskContext
from pydiverse.pipedag.errors import DuplicateNameError
from pydiverse.pipedag.util import normalize_name

if TYPE_CHECKING:
    from pydiverse.pipedag.core.stage import Stage
    from pydiverse.pipedag.materialize.materializing_task import MaterializingTask


@frozen
class Schema:
    """
    Class for holding a schema name with separable prefix and suffix.

    Attributes
    ----------
    name : str
        The schema name.
    prefix : str
        The prefix to be added to the schema name.
    suffix : str
        The suffix to be added to the schema name.
    """

    name: str
    prefix: str = ""
    suffix: str = ""

    def get(self) -> str:
        """
        Get the schema name with prefix and suffix.
        """
        return self.prefix + self.name + self.suffix

    def __str__(self):
        return self.get()


@total_ordering
class Table(Generic[T]):
    """Container for storing Tables.

    Used to wrap table objects that get returned from materializing
    tasks. Tables get stored using the table store.

    .. code-block:: python
       :caption: Example: How to return a table from a task.

        @materialize()
        def task():
            df = pd.DataFrame({"x": [0, 1, 2, 3]}
            return Table(df, "name")

    :param obj: The table object to wrap
    :param name: Optional, case-insensitive name. If no name is provided,
        an automatically generated name will be used. To prevent name collisions,
        you can append ``"%%"`` at the end of the name to enable automatic
        name mangling.
    :param primary_key: Optional name of the primary key that should be
        used when materializing this table. Only supported by some table stores.
    :param indexes: Optional list of indexes to create. Each provided index should be
        a list of column names. Only supported by some table stores.
    :param type_map: Optional map of column names to types. Depending on the table
        store this will allow you to control the datatype as which the specified
        columns get materialized.
    :param nullable: List of columns that should be nullable. If nullable is not None,
        all other columns will be non-nullable.
    :param non_nullable: List of columns that should be non-nullable. If non_nullable
        is not None, all other columns will be nullable.
    :param materialization_details: The label of the materialization_details to be used.
        Overwrites the label given by the stage.

    .. seealso:: You can specify which types of objects should automatically get
        converted to tables using the :ref:`auto_table` config option.
    """

    def __init__(
        self,
        obj: T | None = None,
        name: str | None = None,
        *,
        primary_key: str | list[str] | None = None,
        indexes: list[list[str]] | None = None,
        type_map: dict[str, Any] | None = None,
        nullable: list[str] | None = None,
        non_nullable: list[str] | None = None,
        materialization_details: str | None = None,
        annotation: type | None = None,
    ):
        # state
        self._name = None
        self.stage: Stage | None = None
        self.external_schema: str | None = None
        self.shared_lock_allowed: bool = True

        # arguments
        self.obj = obj
        self.name = name
        self.primary_key = primary_key
        self.indexes = indexes
        self.type_map = type_map
        self.nullable = nullable
        self.non_nullable = non_nullable
        self.materialization_details = materialization_details
        self.annotation = annotation  # if None, will be set by
        #                               core.py:attach_annotation() and used by
        #                               dematerialization hook

        # Check that indexes is of type list[list[str]]
        indexes_type_error = TypeError(
            "Table argument 'indexes' must be of type list[list[str]]. "
            "Make sure you provide a 2d list, not just a 1d list."
        )
        if self.indexes is not None:
            if not isinstance(self.indexes, (list, tuple)):
                raise indexes_type_error
            for index in self.indexes:
                if not isinstance(index, (list, tuple)):
                    raise indexes_type_error
                for col in index:
                    if not isinstance(col, str):
                        raise indexes_type_error
        for arg, name in [
            (self.nullable, "nullable"),
            (self.non_nullable, "non_nullable"),
        ]:
            type_error = TypeError(f"Table argument '{name}' must be of type list[str].")
            if arg is not None:
                if not isinstance(arg, Iterable) or isinstance(arg, str):
                    raise type_error
                if not all(isinstance(x, str) for x in arg):
                    raise type_error

        # ExternalTableReference can reference a table from an external schema
        if isinstance(self.obj, ExternalTableReference):
            self.external_schema = self.obj.schema
            if self.name is not None:
                raise ValueError(
                    "When using an ExternalTableReference, the name of the Table must "
                    "be set via the ExternalTableReference."
                )
            self.name = self.obj.name
            self.shared_lock_allowed = self.obj.shared_lock_allowed

        # cache_key will be overridden shortly before handing over to downstream tasks
        # that use it to compute their input_hash for cache_invalidation due to input
        # change
        self.cache_key = None
        # assumed dependencies are filled by imperative materialization to ensure
        # correct cache invalidation
        self.assumed_dependencies: list[Table] | None = None

    def __repr__(self):
        stage_name = self.stage.name if self.stage else None
        return f"<Table '{self.name}' ({stage_name})>"

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if value is not None and not isinstance(value, str):
            raise TypeError(f"Table name must be of instance 'str' not {type(value)}.")
        self._name = normalize_name(value)

    def copy_without_obj(self) -> "Table":
        obj = self.obj
        self.obj = None
        self_copy = copy.deepcopy(self)
        self.obj = obj
        return self_copy

    def materialize(
        self,
        config_context: ConfigContext | None = None,
        schema: Schema | None = None,
        return_as_type=None,
        return_nothing=False,
    ):
        """Materialize the table.

        This method is used to materialize the table in the table store.
        This imperative way of triggering materialization is useful for debugging
        and it allows materializing subqueries within a task. Every call of this
        function is registered in TaskContext and if any object returned
        by this is returned from an ``@materialize`` task, it will be replaced by
        the respective Table object that was materialized.

        :param config_context: The config context to use for materialization during
            interactive debugging. If it is provided while task is regularly executed,
            it must be identical to ConfigContext.get().
        :param return_as_type: By default, the materialized table is returned as the
            input_type of the task (for dataframes with AUTO_VERSION support, it
            just dematerializes the empty dataframes). With return_as you can
            override this behavior and explicitly choose the dematiarlization type.
            ``return_as=sqlalchemy.Table`` might be a useful choice for tasks with
            pandas or polars input_type. It is also possible to pass an `Iterable` like
            list in order to receive a tuple of the dematerialized objects.
        :param return_nothing: If True, the method will return None instead of the
            dematerialized created table.
        """
        try:
            task_context = TaskContext.get()  # raises Lookup Error if no TaskContext is open
            if config_context is not None and config_context is not ConfigContext.get():
                raise ValueError(
                    "config_context must be identical to ConfigContext.get() "
                    "when task is regularly executed by pipedag orchestration."
                )
            config_context = ConfigContext.get()
            task_schema = config_context.store.table_store.get_schema(task_context.task.stage.transaction_name)
            if schema is not None and schema != task_schema:
                raise ValueError(
                    "schema must be identical to Task Stage transaction schema "
                    "when task is regularly executed by pipedag orchestration."
                )
            schema = task_schema
            if task_context.imperative_materialize_callback is None:
                # this path is triggered while determining auto-version
                return self.obj
            try:
                return task_context.imperative_materialize_callback(
                    self, config_context, return_as_type, return_nothing
                )
            except (RuntimeError, DuplicateNameError):
                # fall back to debug materialization when Table.materialize() is
                # called twice for the same table
                task_context.task.logger.info(
                    "Falling back to debug materialization due to duplicate materializtion of this table"
                )

                def return_type_mutator(return_as_type):
                    if return_as_type is None:
                        task: MaterializingTask = task_context.task  # type: ignore
                        return_as_type = task.input_type
                        if return_as_type is None or not config_context.store.table_store.get_r_table_hook(
                            return_as_type
                        ).retrieve_as_reference(return_as_type):
                            # dematerialize as sa.Table if it would transfer all rows
                            # to python when dematerializing with input_type
                            return_as_type = sa.Table
                    return return_as_type

                if isinstance(return_as_type, Iterable):
                    return_as_type = tuple(return_type_mutator(t) for t in return_as_type)
                else:
                    return_as_type = return_type_mutator(return_as_type)
        except LookupError:
            # LookupError happens if no TaskContext is open
            pass
        if config_context is not None:
            if schema is None:
                raise ValueError(
                    "schema must be provided when task is not regularly executed by pipedag orchestration."
                )
            # fall back to debug behavior when an explicit table_store is given
            # via config_context
            from pydiverse.pipedag.materialize import debug

            table_name = debug.materialize_table(
                table=self,
                config_context=config_context,
                schema=schema,
            )
            if not return_nothing:

                def get_return_obj(return_as_type):
                    if return_as_type is None:
                        # use sqlalchemy as default reference dematerialization
                        return_as_type = sa.Table
                    store = config_context.store.table_store
                    hook = store.get_r_table_hook(return_as_type)
                    save_name = self.name
                    self.name = table_name
                    schema_name = schema.name if store.get_schema(schema.name).get() == schema.get() else schema.get()
                    if store.get_schema(schema_name).get() != schema.get():
                        raise ValueError(
                            "Schema prefix and postfix must match prefix and postfix of"
                            " provided config_context: "
                            f"{store.get_schema(schema_name).get()} != {schema.get()}"
                        )
                    obj = hook.retrieve(
                        config_context.store.table_store,
                        self,
                        schema_name,
                        return_as_type,
                    )
                    self.name = save_name
                    return obj

                if isinstance(return_as_type, Iterable):
                    return tuple(get_return_obj(t) for t in return_as_type)
                else:
                    return get_return_obj(return_as_type)

        elif not return_nothing:
            # We support calling dataframe tasks outside flow declaration by
            # assuming that input tables are already given dematerialized by the
            # caller and returning dataframe objects unmaterialized. In this scenario,
            # we still like `return Table(...)` to behave identical to
            # `return Table(...).materialize()` at the end of tasks.
            # Furthermore, we can support
            # `df = Table(...).materialize(return_as_type=pd.DataFrame)`
            if return_as_type is not None:
                logger = structlog.get_logger(self.__class__.__name__, table=self)
                logger.info(
                    "Ignoring return_as_type in Table.materialize() outside of flow without given config_context.",
                    return_as_type=return_as_type,
                )
            return self.obj

    def __getstate__(self):
        # The table `obj` field can't necessarily be pickled. That's why we remove it
        # from the state before pickling.
        state = self.__dict__.copy()
        state["obj"] = None
        return state

    def __lt__(self, other):
        if self.stage is not None or other.stage is not None:
            if self.stage < other.stage:
                return True
            if self.stage > other.stage:
                return False
        return self.name < other.name

    def __eq__(self, other: "Table"):
        if not isinstance(other, Table):
            return False
        return self.name == other.name and (self.stage == other.stage or (self.stage is None and other.stage is None))

    def __hash__(self):
        return hash(self.name) if self.stage is None else hash((self.name, self.stage))


class RawSql:
    """Container for raw sql strings.

    This allows returning sql query strings that then get executed in the
    table store. This is only intended to help with transitioning legacy sql
    pipelines to pipedag, and should be replaced with pipedag managed tables as
    soon as possible.

    .. attention::
        When using RawSql, make sure that you only write tables to the stage that
        the corresponding task is running in. Otherwise, schema swapping won't work.
        To do this, pass the current stage as an argument to your task and then
        access the current stage name using :py:class:`Stage.current_name`.

    :param sql: The sql query string to execute. Depending on the database dialect
        and the separator parameter, the query will be split into multiple subqueries
        that then get executed sequentially.
    :param name: Optional, case-insensitive name. If no name is provided,
        an automatically generated name will be used. To prevent name collisions,
        you can append ``"%%"`` at the end of the name to enable automatic
        name mangling.
    :param separator: The separator used when splitting the query into subqueries.
        Default: ``";"``

    Example
    -------
    When you want to use tables produced by a RawSql task in a downstream task,
    you can either use square brackets during flow definition to pass
    a specific table to a child task, or you can pass the entire RawSql object
    as an input, in which case all tables get loaded and the child task can
    access them using square brackets (or any of the dict-like methods
    defined by RawSql).

    .. code-block::
        :emphasize-lines: 23-24, 26-27

        @materialize(lazy=True)
        def raw_sql_task(stage):
            schema = stage.current_name
            return RawSql(f\"\"\"
                CREATE TABLE {schema}.tbl_1 AS SELECT 1 as x;
                CREATE TABLE {schema}.tbl_2 AS SELECT 2 as x;
            \"\"\")

        @materialize(input_type=sa.Table)
        def foo(tbl_1, tbl_2):
            ...

        @materialize(input_type=sa.Table)
        def bar(raw_sql):
            tbl_1 = raw_sql["tbl_1"]
            tbl_2 = raw_sql["tbl_2"]
            ...

        with Flow() as f:
            with Stage("stage") as s:
                out = raw_sql_task(s)

                # Manually pass specific tables into the task
                foo(out["tbl_1"], out["tbl_2"])

                # Or pass the entire RawSql object into the task
                bar(out)
    """

    def __init__(
        self,
        sql: str | None = None,
        name: str | None = None,
        separator: str = ";",
    ):
        self._name = None
        self.stage: Stage | None = None

        self.sql = sql
        self.name = name
        self.separator = separator

        # cache_key will be overridden shortly before handing over to downstream tasks
        # that use it to compute their input_hash for cache_invalidation due to input
        # change
        self.cache_key = None

        # assumed dependencies are filled by imperative materialization to ensure
        # correct cache invalidation
        self.assumed_dependencies: list[Table] | None = None

        # If a task receives a RawSQL object as input, it loads all tables
        # produced by it and makes them available through a dict like interface.
        self.table_names: list[str] | None = None  # type: ignore
        self.loaded_tables: dict[str, Any] | None = None  # type: ignore

    def __repr__(self):
        stage_name = self.stage.name if self.stage else None
        sql_short = None
        if self.sql:
            sql_short = self.sql.strip()[0:40].replace("\n", "").strip()
        return f"<Raw SQL '{self.name}' ({stage_name}) - {sql_short}>"

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if value is not None and not isinstance(value, str):
            raise TypeError(f"RawSql name must be of instance 'str' not {type(value)}.")
        self._name = normalize_name(value)

    # Dict-like interface

    def __iter__(self) -> Iterable[str]:
        """Yields all names of tables produced by this RawSql object."""
        yield from (self.table_names or [])

    def __contains__(self, table_name: str) -> bool:
        """Check if this RawSql object produced a table with name `table_name`."""
        return table_name in (self.table_names or [])

    def __getitem__(self, table_name: str):
        """Gets the table produced by this RawSql object with name `table_name`."""
        if table_name not in (self.table_names or []):
            raise KeyError(f"No table with name '{table_name}' found in RawSql.")

        # Did load tables -> __getitem__ should return Table.obj
        if self.loaded_tables is not None:
            return self.loaded_tables[table_name]

        # Otherwise, while preparing the task inputs inside
        # (`TaskGetItem.resolve_value`) we need to return objects of type `Table`.
        table = Table(None, table_name)
        table.stage = self.stage
        table.cache_key = self.cache_key
        return table

    def items(self) -> Iterable[tuple[str, Any]]:
        """Returns pairs of ``(table_name, table)``."""
        for table_name in self.table_names or []:
            yield table_name, self[table_name]

    def get(self, table_name: str, default=None):
        """
        Returns the table with name `table_name`, or if no such table exists,
        the `default` value.
        """
        if table_name in self:
            return table_name
        return default

    def copy_without_obj(self) -> "RawSql":
        obj = self.loaded_tables
        self.loaded_tables = None
        self_copy = copy.deepcopy(self)
        self.loaded_tables = obj
        return self_copy

    def __getstate__(self):
        # The `loaded_tables` field can't necessarily be pickled. That's why we
        # remove it from the state before pickling.
        state = self.__dict__.copy()
        state["loaded_tables"] = None
        return state


class Blob(Generic[T]):
    """Blob (binary large object) container.

    Used to wrap arbitrary Python objects that get returned from materializing
    tasks. Blobs get stored in the blob store.

    .. code-block:: python
       :caption: Example: How to return a blob from a task.

        @materialize()
        def task():
            obj = SomePicklableClass()
            return Blob(obj, "name")

    :param obj: The object to wrap
    :param name: Optional, case-insensitive name. If no name is provided,
        an automatically generated name will be used. To prevent name collisions,
        you can append ``"%%"`` at the end of the name to enable automatic
        name mangling.

    .. seealso:: You can specify which types of objects should automatically get
        converted to blobs using the :ref:`auto_blob` config option.
    """

    def __init__(
        self,
        obj: T | None = None,
        name: str | None = None,
    ):
        self._name = None
        self.stage: Stage | None = None

        self.obj = obj
        self.name = name

        # cache_key will be overridden shortly before handing over to downstream tasks
        # that use it to compute their input_hash for cache_invalidation due to input
        # change
        self.cache_key = None

    def __repr__(self):
        stage_name = self.stage.name if self.stage else None
        return f"<Blob '{self.name}' ({stage_name})>"

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = normalize_name(value)

    def copy_without_obj(self) -> "Blob":
        obj = self.obj
        self.obj = None
        self_copy = copy.deepcopy(self)
        self.obj = obj
        return self_copy


class ExternalTableReference:
    """Reference to a user-created table.

    By returning a `ExternalTableReference` wrapped in a :py:class:`~.Table` from,
    a task you can tell pipedag about a table, a view or DB2 nickname in an external
    `schema`. The schema may be a multi-part identifier like "[db_name].[schema_name]"
    if the database supports this. It is passed to SQLAlchemy as-is.

    Only supported by :py:class:`~.SQLTableStore`.

    Warning
    -------
    When using a `ExternalTableReference`, pipedag has no way of knowing the cache
    validity of the external object. Hence, the user should provide a cache function
    for the `Task`.
    It is now allowed to specify a `ExternalTableReference` to a table in schema of the
    current stage.

    Example
    -------
    You can use a `ExternalTableReference` to tell pipedag about a table that exists
    in an external schema::

        @materialize(version="1.0")
        def task():
            return Table(ExternalTableReference("name_of_table", "schema"))

    By using a cache function, you can establish the cache (in-)validity of the
    external table::

        from datetime import date

        # The external table becomes cache invalid every day at midnight
        def my_cache_fun():
            return date.today().strftime("%Y/%m/%d")

        @materialize(cache=my_cache_fun)
        def task():
            return Table(ExternalTableReference("name_of_table", "schema"))

    The ExternalTableReference object can also be created at flow wiring time::

        with Flow() as f:
            with Stage("stage") as s:
                tbl = Table(ExternalTableReference("name_of_table", "schema"))
                _ = some_task(tbl)
    """

    def __init__(self, name: str, schema: str, shared_lock_allowed: bool = False):
        """
        :param name: The name of the table, view, or nickname/alias
        :param schema: The external schema of the object. A multi-part schema
            is allowed with '.' separator as also supported by SQLAlchemy Table
            schema argument.
        :param shared_lock_allowed: Whether to disable acquiring a shared lock
            when using the object in a SQL query. If set to `False`, no lock is
            used. This is useful when the user is not allowed to lock the table.
            If pipedag does not lock source tables for this dialect, this argument
            has no effect. The default is `False`.
        """
        self.name = name
        self.schema = schema
        self.shared_lock_allowed = shared_lock_allowed

    def __repr__(self):
        return f"<ExternalTableReference: {hex(id(self))} (schema: {self.schema})>"


def attach_annotation(annotation: type, arg):
    """Recursive traversal for attaching type annotations to correct container."""
    if typing.get_origin(annotation) is not None:
        anno_origin = typing.get_origin(annotation)
        anno_args = typing.get_args(annotation)
        if (
            inspect.isclass(anno_origin)
            and issubclass(anno_origin, dict)
            and len(anno_args) == 2
            and isinstance(arg, dict)
        ):
            for key, value in arg.items():
                attach_annotation(anno_args[0], key)
                attach_annotation(anno_args[1], value)
        elif (
            inspect.isclass(anno_origin)
            and issubclass(anno_origin, list)
            and len(anno_args) == 1
            and isinstance(arg, Iterable)
        ):
            for value in arg:
                attach_annotation(anno_args[0], value)
        elif (
            inspect.isclass(anno_origin)
            and issubclass(anno_origin, tuple)
            and isinstance(arg, typing.Sized)
            and len(anno_args) == len(arg)
        ):
            for value, anno in zip(arg, anno_args):
                attach_annotation(anno, value)
    if isinstance(arg, (Table, Blob, RawSql)):
        arg.annotation = annotation


try:
    import pydiverse.transform as pdt

    @overload
    def materialize_table(name: str | None = None, table_prefix: str | None = None): ...

    @pdt.verb
    def materialize_table(tbl: pdt.Table, name: str | None = None, table_prefix: str | None = None):
        # use imperative materialization of pipedag within pydiverse transform task
        if name is None:
            name = table_prefix or ""
            name += tbl._ast.name or ""
            name += "%%"
        if tbl >> pdt.build_query():
            return Table(tbl, name=name).materialize(return_as_type=pdt.SqlAlchemy)
        return tbl >> pdt.alias()
except ImportError:
    # If pydiverse.transform is not available, we cannot use the materialize_table verb
    def materialize_table(tbl: Any, table_prefix: str | None = None):
        raise ImportError(
            "pydiverse.transform is not available. Please install pydiverse-transform to use materialize_table."
        )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pandas as pd
import pytest
import sqlalchemy as sa

from pydiverse.pipedag import AUTO_VERSION, Blob, ConfigContext, Flow, Stage, Table
from pydiverse.pipedag.container import RawSql
from pydiverse.pipedag.context import StageLockContext
from pydiverse.pipedag.context.context import CacheValidationMode
from pydiverse.pipedag.materialize.core import (
    input_stage_versions,
    materialize,
)
from pydiverse.pipedag.util.sql import compile_sql

# Parameterize all tests in this file with several instance_id configurations
from tests.fixtures.instances import ALL_INSTANCES, skip_instances, with_instances
from tests.util import select_as
from tests.util import tasks_library as m
from tests.util import tasks_library_imperative as m2
from tests.util.spy import spy_task
from tests.util.tasks_library import get_task_logger

try:
    import polars as pl
except ImportError:
    pl = None

# snowflake tests are too slow, possibly they could move to nightly tests
pytestmark = [with_instances(tuple(set(ALL_INSTANCES) - {"snowflake"}))]


# Test Basic Cache Invalidation Behaviour


def test_change_bound_argument(mocker):
    input_list = [1]

    with Flow() as flow:
        with Stage("stage_1"):
            out = m.noop(input_list)
            child = m.noop2(out)

    # Initial Call
    with StageLockContext():
        result = flow.run()
        assert result.get(out)[0] == 1
        assert result.get(child)[0] == 1

    # Calling flow.run again shouldn't call the task
    out_spy = spy_task(mocker, out)
    child_spy = spy_task(mocker, child)
    with StageLockContext():
        result = flow.run()
        assert result.get(out)[0] == 1
        assert result.get(child)[0] == 1
        out_spy.assert_not_called()
        child_spy.assert_not_called()

    # Changing the input object should invalidate the cache
    input_list[0] = 2
    with StageLockContext():
        result = flow.run()
        assert result.get(out)[0] == 2
        assert result.get(child)[0] == 2
        out_spy.assert_called_once()
        child_spy.assert_called_once()


def test_changed_cache_fn_literal(mocker):
    cache_value = 0

    def cache():
        return cache_value

    @materialize(cache=cache, version="1.0")
    def return_cache_value():
        return cache_value

    with Flow() as flow:
        with Stage("stage_1"):
            out = return_cache_value()
            child = m.noop(out)

    # Initial Call
    with StageLockContext():
        result = flow.run()
        assert result.get(out) == 0
        assert result.get(child) == 0

    # Calling flow.run again shouldn't call the task
    out_spy = spy_task(mocker, out)
    child_spy = spy_task(mocker, child)
    with StageLockContext():
        result = flow.run()
        assert result.get(out) == 0
        assert result.get(child) == 0
        out_spy.assert_not_called()
        child_spy.assert_not_called()

    # Changing the cache value should cause it to get called again
    cache_value = 1
    with StageLockContext():
        result = flow.run()
        assert result.get(out) == 1
        assert result.get(child) == 1
        out_spy.assert_called_once()
        child_spy.assert_called_once()


@pytest.mark.parametrize("imperative", [False, True])
def test_change_cache_fn_table(mocker, imperative):
    cache_value = 0

    def cache():
        return cache_value

    @materialize(cache=cache, version="1.0")
    def return_cache_table():
        tbl = Table(select_as(cache_value, "x"))
        return tbl.materialize() if imperative else tbl

    @materialize(input_type=pd.DataFrame, version="1.0")
    def get_first(table, col):
        return int(table[col][0])

    with Flow() as flow:
        with Stage("stage_1"):
            out = return_cache_table()
            child = get_first(out, "x")

    # Initial Call
    with StageLockContext():
        result = flow.run()
        assert result.get(child) == 0

    # Calling flow.run again shouldn't call the task
    out_spy = spy_task(mocker, out)
    child_spy = spy_task(mocker, child)
    with StageLockContext():
        result = flow.run()
        assert result.get(child) == 0
        out_spy.assert_not_called()
        child_spy.assert_not_called()

    # Changing the cache value should cause it to get called again
    cache_value = 1
    with StageLockContext():
        result = flow.run()
        assert result.get(child) == 1
        out_spy.assert_called_once()
        child_spy.assert_called_once()


@with_instances("postgres")
def test_change_cache_fn_blob(mocker):
    cache_value = 0

    def cache():
        return cache_value

    @materialize(cache=cache, version="1.0")
    def return_cache_blob():
        return Blob(cache_value)

    with Flow() as flow:
        with Stage("stage_1"):
            out = return_cache_blob()
            child = m.noop(out)

    # Initial Call
    with StageLockContext():
        result = flow.run()
        assert result.get(out) == 0
        assert result.get(child) == 0

    # Calling flow.run again shouldn't call the task
    out_spy = spy_task(mocker, out)
    child_spy = spy_task(mocker, child)
    with StageLockContext():
        result = flow.run()
        assert result.get(out) == 0
        assert result.get(child) == 0
        out_spy.assert_not_called()
        child_spy.assert_not_called()

    # Changing the cache value should cause it to get called again
    cache_value = 1
    with StageLockContext():
        result = flow.run()
        assert result.get(out) == 1
        assert result.get(child) == 1
        out_spy.assert_called_once()
        child_spy.assert_called_once()


def test_change_task_version_literal(mocker):
    with Flow() as flow:
        with Stage("stage_1"):
            out = m.one()
            child = m.noop(out)

    # Initial Call
    out._version = "VERSION 0"
    assert flow.run().successful

    # Second Call (Should be cached)
    out_spy = spy_task(mocker, out)
    child_spy = spy_task(mocker, child)
    assert flow.run().successful
    out_spy.assert_not_called()
    child_spy.assert_not_called()

    # Changing the version should invalidate the cache, but the child still
    # shouldn't get called because the parent task still returned the same value.
    out._version = "VERSION 1"
    assert flow.run().successful
    out_spy.assert_called_once()
    child_spy.assert_not_called()


def test_change_task_version_table(mocker):
    with Flow() as flow:
        with Stage("stage_1"):
            out = m.simple_dataframe()
            child = m.noop(out)
            child2 = m.noop_sql(out)  # lazy=False task
            child_lazy = m.noop_lazy(out)  # lazy=True task

    # Initial Call
    out._version = "VERSION 0"
    assert flow.run().successful

    # Second Call (Should be cached)
    out_spy = spy_task(mocker, out)
    child_spy = spy_task(mocker, child)
    child2_spy = spy_task(mocker, child2)
    child_lazy_spy = spy_task(mocker, child_lazy)
    assert flow.run().successful
    out_spy.assert_not_called()
    child_spy.assert_not_called()
    child2_spy.assert_not_called()
    child_lazy_spy.assert_called_once()

    # Changing the version should invalidate the cache. This should also invalidate
    # the child task because it receives the table as input.
    out._version = "VERSION 1"
    assert flow.run().successful
    out_spy.assert_called_once()
    child_spy.assert_called_once()
    child2_spy.assert_called_once()
    child_lazy_spy.assert_called_once()


@with_instances("postgres")
def test_change_task_version_blob(mocker):
    with Flow() as flow:
        with Stage("stage_1"):
            out = m.object_blob({"x": "y"})
            child = m.as_blob(out)

    # Initial Call
    out._version = "VERSION 0"
    assert flow.run().successful

    # Second Call (Should be cached)
    out_spy = spy_task(mocker, out)
    child_spy = spy_task(mocker, child)
    assert flow.run().successful
    out_spy.assert_not_called()
    child_spy.assert_not_called()

    # Changing the version should invalidate the cache. This should also invalidate
    # the child task because it receives the blob as input.
    out._version = "VERSION 1"
    assert flow.run().successful
    out_spy.assert_called_once()
    child_spy.assert_called_once()


def test_change_lazy_query(mocker):
    query_value = 1

    @materialize(lazy=True, nout=2)
    def lazy_task():
        return 0, Table(select_as(query_value, "x"), name="lazy_table")

    @materialize(input_type=pd.DataFrame, version="1.0")
    def get_first(table, col):
        return int(table[col][0])

    with Flow() as flow:
        with Stage("stage_1"):
            const, lazy = lazy_task()
            value = get_first(lazy, "x")
            const2 = m.noop(const)
            lazy2 = m.noop_sql(lazy)  # lazy=False task
            lazy3 = m.noop_lazy(lazy)

    # Initial Run
    lazy_spy = spy_task(mocker, lazy)
    with StageLockContext():
        result = flow.run()
        assert result.successful
        assert result.get(value) == 1
        lazy_spy.assert_called_once()

    # Second run, because the task is lazy, it should always get called.
    # The value task however shouldn't get called.
    value_spy = spy_task(mocker, value)
    const2_spy = spy_task(mocker, const2)
    lazy2_spy = spy_task(mocker, lazy2)
    lazy3_spy = spy_task(mocker, lazy3)
    with StageLockContext():
        result = flow.run()
        assert result.successful
        assert result.get(value) == 1
        lazy_spy.assert_called_once()
        value_spy.assert_not_called()
        const2_spy.assert_not_called()
        lazy2_spy.assert_not_called()
        lazy3_spy.assert_called_once()

    # Third run with changed query_value
    query_value = 2
    with StageLockContext():
        result = flow.run()
        assert result.successful
        assert result.get(value) == 2
        lazy_spy.assert_called_once()
        value_spy.assert_called_once()
        const2_spy.assert_not_called()
        lazy2_spy.assert_called_once()
        lazy3_spy.assert_called_once()


def test_change_raw_sql(mocker):
    query_value = 1

    @materialize(lazy=True, nout=2, version="1.0")
    def raw_task(stage):
        raw_sql = compile_sql(select_as(query_value, "x"))
        return 0, RawSql(raw_sql, "raw_task")

    @materialize(version="1.0")
    def raw_child(raw):
        return raw.sql

    with Flow() as flow:
        with Stage("stage_1") as s:
            const, raw = raw_task(s)
            child = raw_child(raw)
            const = m.noop(const)

    # Initial Run
    raw_spy = spy_task(mocker, raw)
    with StageLockContext():
        result = flow.run()
        assert result.successful
        raw_spy.assert_called_once()

    # Second run, because the task is lazy, it should always get called.
    # The value task however shouldn't get called.
    child_spy = spy_task(mocker, child)
    const_spy = spy_task(mocker, const)
    with StageLockContext():
        result = flow.run()
        assert result.successful
        assert "SELECT 1" in result.get(child).upper()
        raw_spy.assert_called_once()
        child_spy.assert_not_called()
        const_spy.assert_not_called()

    # Third run with changed query_str
    query_value = 2
    with StageLockContext():
        result = flow.run()
        assert result.successful
        assert "SELECT 2" in result.get(child).upper()
        raw_spy.assert_called_once()
        child_spy.assert_called_once()
        const_spy.assert_not_called()


def test_change_task_stage_literal(mocker):
    with Flow() as flow:
        with Stage("stage_1"):
            one = m.one()
        with Stage("stage_2"):
            m.noop(0)  # This is to clear the stage
        with Stage("stage_3"):
            child = m.noop(one)
    _ = child

    assert flow.run().successful

    with Flow() as flow:
        with Stage("stage_2"):
            one = m.one()
        with Stage("stage_3"):
            child = m.noop(one)

    one_spy = spy_task(mocker, one)
    child_spy = spy_task(mocker, child)

    # Moving the one task to a different stage should cause it to be called again,
    # but because its return value is the same, the child task shouldn't get called.
    assert flow.run().successful
    one_spy.assert_called_once()
    child_spy.assert_not_called()


def test_change_task_stage_table(mocker):
    with Flow() as flow:
        with Stage("stage_1"):
            table = m.simple_dataframe()
        with Stage("stage_2"):
            m.noop(0)  # This is to clear the stage
        with Stage("stage_3"):
            child = m.noop(table)

    assert flow.run().successful

    with Flow() as flow:
        with Stage("stage_2"):
            table = m.simple_dataframe()
        with Stage("stage_3"):
            child = m.noop(table)

    table_spy = spy_task(mocker, table)
    child_spy = spy_task(mocker, child)

    # Moving the table task to a different stage should cause it to be called again,
    # and because it returns a table, any child task should also get invalidated.
    assert flow.run().successful
    table_spy.assert_called_once()
    child_spy.assert_called_once()


@with_instances("postgres")
def test_change_task_stage_blob(mocker):
    with Flow() as flow:
        with Stage("stage_1"):
            blob = m.as_blob(1)
        with Stage("stage_2"):
            m.noop(0)  # This is to clear the stage
        with Stage("stage_3"):
            child = m.noop(blob)

    assert flow.run().successful

    with Flow() as flow:
        with Stage("stage_2"):
            blob = m.as_blob(1)
        with Stage("stage_3"):
            child = m.noop(blob)

    blob_spy = spy_task(mocker, blob)
    child_spy = spy_task(mocker, child)

    # Moving the blob task to a different stage should cause it to be called again,
    # and because it returns a blob, any child task should also get invalidated.
    assert flow.run().successful
    blob_spy.assert_called_once()
    child_spy.assert_called_once()


def test_different_task_same_input(mocker):
    with Flow() as flow:
        with Stage("stage_1"):
            one = m.one()
            two = m.two()

    # Initial run
    with StageLockContext():
        result = flow.run()
        assert result.successful
        assert result.get(one) == 1
        assert result.get(two) == 2

    # Second Run should be cached
    one_spy = spy_task(mocker, one)
    two_spy = spy_task(mocker, two)

    with StageLockContext():
        result = flow.run()
        assert result.successful
        assert result.get(one) == 1
        assert result.get(two) == 2
        one_spy.assert_not_called()
        two_spy.assert_not_called()


def test_same_task_different_stages(mocker):
    with Flow() as flow:
        with Stage("stage_1"):
            out_s1 = m.one()
        with Stage("stage_2"):
            out_s2 = m.one()

    # Initial run
    with StageLockContext():
        result = flow.run()
        assert result.successful
        assert result.get(out_s1) == 1
        assert result.get(out_s2) == 1

    # Second Run should be cached
    out_spy = spy_task(mocker, out_s1)

    with StageLockContext():
        result = flow.run()
        assert result.successful
        assert result.get(out_s1) == 1
        assert result.get(out_s2) == 1
        out_spy.assert_not_called()


def test_change_version_table(mocker):
    cache_value = 0

    def cache():
        return cache_value

    @materialize(lazy=True, cache=cache)
    def return_cache_table():
        return Table(select_as(cache_value, "x"))

    def get_flow(version):
        @materialize(version=version, input_type=pd.DataFrame)
        def named_copy(table: pd.DataFrame):
            return Table(table, "_table_copy")

        with Flow() as flow:
            with Stage("stage_1"):
                out = return_cache_table()
                cpy = named_copy(out)
        return flow, out, cpy

    flow, out, cpy = get_flow(version="1.0")
    # Initial Call
    with StageLockContext():
        result = flow.run()
        assert result.get(cpy)["x"].iloc[0] == 0

    # Calling flow.run again shouldn't call the task
    out_spy = spy_task(mocker, out)
    cpy_spy = spy_task(mocker, cpy)
    with StageLockContext():
        result = flow.run()
        assert result.get(cpy)["x"].iloc[0] == 0
        out_spy.assert_called_once()  # lazy task is always called
        cpy_spy.assert_not_called()

    # Changing the cache value should cause it to get called again
    cache_value = 1
    with StageLockContext():
        result = flow.run()
        assert result.get(cpy)["x"].iloc[0] == 1
        out_spy.assert_called_once()
        cpy_spy.assert_called_once()

    # recreating flow without version change should not invalidate cache
    flow, out, cpy = get_flow(version="1.0")
    out_spy = spy_task(mocker, out)
    cpy_spy = spy_task(mocker, cpy)
    with StageLockContext():
        result = flow.run()
        assert result.get(cpy)["x"].iloc[0] == 1
        out_spy.assert_called_once()  # lazy task is always called
        cpy_spy.assert_not_called()

    # Changing the version value should cause cpy to get recreated again
    flow, out, cpy = get_flow(version="1.1")
    out_spy = spy_task(mocker, out)
    cpy_spy = spy_task(mocker, cpy)
    with StageLockContext():
        result = flow.run()
        assert result.get(cpy)["x"].iloc[0] == 1
        out_spy.assert_called_once()  # lazy task is always called
        cpy_spy.assert_called_once()

    # Changing the version value should cause cpy to get recreated again
    for _ in range(3):
        flow, out, cpy = get_flow(version=None)
        out_spy = spy_task(mocker, out)
        cpy_spy = spy_task(mocker, cpy)
        with StageLockContext():
            result = flow.run()
            assert result.get(cpy)["x"].iloc[0] == 1
            out_spy.assert_called_once()  # lazy task is always called
            cpy_spy.assert_called_once()


@pytest.mark.parametrize("ignore_task_version", [True, False])
@pytest.mark.parametrize("disable_cache_function", [True, False])
@pytest.mark.parametrize("mode", ["ASSERT_NO_FRESH_INPUT", "NORMAL"])
def test_cache_validation_mode_assert(ignore_task_version, disable_cache_function, mode):
    mode = getattr(CacheValidationMode, mode.upper())
    kwargs = dict(
        ignore_task_version=ignore_task_version,
        disable_cache_function=disable_cache_function,
        cache_validation_mode=mode,
    )
    if not disable_cache_function and mode == CacheValidationMode.NORMAL:
        pytest.skip("Combination does not provoke exception")
        return
    error = AssertionError if mode == CacheValidationMode.ASSERT_NO_FRESH_INPUT else ValueError

    cache_value = 0
    return_value = 0

    def cache():
        return cache_value

    def get_flow(version):
        @materialize(version=version, cache=cache)
        def source():
            return pd.DataFrame(dict(x=[return_value]))

        @materialize(lazy=True, cache=cache)
        def lazy_source():
            return Table(select_as(return_value, "x"))

        with Flow() as flow:
            with Stage("stage_1"):
                s1 = source()
                s2 = lazy_source()
        return flow, s1, s2

    flow, s1, s2 = get_flow(version="1.0")
    # Initial Call
    with StageLockContext():
        result = flow.run(
            cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID,
            disable_cache_function=False,
            ignore_task_version=False,
        )
        assert result.successful
        assert result.get(s1, as_type=pd.DataFrame)["x"].iloc[0] == return_value
        assert result.get(s2, as_type=pd.DataFrame)["x"].iloc[0] == return_value

    if disable_cache_function or ignore_task_version:
        with pytest.raises(error):
            result = flow.run(**kwargs)
            assert result.successful

    cache_value = 1  # IGNORE_FRESH_INPUT should be implied
    if not disable_cache_function and not ignore_task_version and mode != CacheValidationMode.NORMAL:
        result = flow.run(
            cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID,
            disable_cache_function=False,
            ignore_task_version=False,
        )
        assert result.successful

    return_value = 1
    with pytest.raises(error):
        flow.run(**kwargs)

    return_value = 0
    flow, _, _ = get_flow(version="1.1")
    with pytest.raises(error):
        flow.run(**kwargs)


try:
    import dask
except ImportError:
    dask = None


@with_instances("postgres", "local_table_cache")
@pytest.mark.skipif(dask is None, reason="AUTO_VERSION for pandas requires dask")
@pytest.mark.parametrize("ignore_task_version", [True, False])
@pytest.mark.parametrize("disable_cache_function", [True, False])
@pytest.mark.parametrize("mode", ["NORMAL", "IGNORE_FRESH_INPUT", "FORCE_FRESH_INPUT", "FORCE_CACHE_INVALID"])
@pytest.mark.parametrize("imperative", [False, True])
def test_cache_validation_mode(ignore_task_version, disable_cache_function, mode, imperative, mocker):
    _m = m2 if imperative else m
    mode = getattr(CacheValidationMode, mode.upper())
    if disable_cache_function and mode == CacheValidationMode.NORMAL:
        pytest.skip("Cannot disable cache function in mode NORMAL")
        return

    kwargs = dict(
        ignore_task_version=ignore_task_version,
        disable_cache_function=disable_cache_function,
        cache_validation_mode=mode,
    )
    cache_calls = [0]  # consider using mocker also for this
    cache_value = 0

    def cache():
        cache_calls[0] += 1
        return cache_value

    def cache2(df):
        cache_calls[0] += 1
        return cache_value

    @materialize(lazy=True, cache=cache)
    def return_cache_table_lazy():
        tbl = Table(select_as(cache_value, "x"))
        return tbl.materialize() if imperative else tbl

    @materialize(version=None, cache=cache)
    def return_cache_table_always():
        df = pd.DataFrame(dict(x=[cache_value]))
        return Table(df).materialize() if imperative else df

    @materialize(version=AUTO_VERSION, cache=cache2, input_type=pd.DataFrame)
    def return_cache_table_auto(df: pd.DataFrame):
        return Table(df).materialize() if imperative else df

    @materialize(version="1.0", cache=cache)
    def return_cache_table():
        df = pd.DataFrame(dict(x=[cache_value]))
        return Table(df).materialize() if imperative else df

    @input_stage_versions(lazy=True, input_type=sa.Table)
    def dummy_copy_inputs(
        transaction: dict[str, sa.sql.selectable.Alias],
        other: dict[str, sa.sql.selectable.Alias],
    ):
        _ = other  # we cannot make any assumptions on the other stage version
        assert len(transaction) == 0
        return Table(sa.select(sa.literal(1).label("a")), name="dummy")

    @input_stage_versions(lazy=True, input_type=sa.Table)
    def validate_stage(
        transaction: dict[str, sa.sql.selectable.Alias],
        other: dict[str, sa.sql.selectable.Alias],
    ):
        _ = other  # we cannot make any assumptions on the other stage version
        get_task_logger().info(f"Transaction tables: {transaction}")
        transaction_schema = {tbl.original.schema for tbl in transaction.values()} - {
            tbl.original.schema for tbl in other.values()
        }
        assert transaction_schema
        assert transaction_schema.pop().startswith({tbl.original.schema for tbl in other.values()}.pop().split("__")[0])
        assert len([tbl for tbl in transaction.keys() if not tbl.endswith("__copy")]) == 12

    @input_stage_versions(lazy=True, input_type=sa.Table)
    def validate_stage2(
        transaction: dict[str, sa.sql.selectable.Alias],
        other: dict[str, sa.sql.selectable.Alias],
    ):
        # it is expected that we have a "Failed to retrieve"-exception in other stage
        _ = other
        assert len(transaction) == 1

    def get_flow():
        with Flow() as flow:
            with Stage("stage_1"):
                x = dummy_copy_inputs()
                _ = m.noop(x)
                out_lazy = return_cache_table_lazy()
                out_always = return_cache_table_always()
                out_auto = return_cache_table_auto(out_lazy)
                out = return_cache_table()
                outs = [out_lazy, out_always, out_auto, out]
                cpy = [_m.noop(o) for o in outs]
                ind1 = _m.simple_dataframe()
                ind2 = _m.simple_lazy_table()
                validate_stage()
                validate_stage2(ind2)
        return flow, outs, cpy, [ind1, ind2]

    flow, outs, cpy, ind = get_flow()
    # Initial Call
    with StageLockContext():
        result = flow.run(
            cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID,
            disable_cache_function=False,
            ignore_task_version=False,
        )
        assert all(result.get(c, as_type=pd.DataFrame)["x"].iloc[0] == 0 for c in cpy)
    assert cache_calls[0] == 4
    cache_calls[0] = 0

    # Calling flow.run again
    out_spy = [spy_task(mocker, o) for o in outs]
    cpy_spy = [spy_task(mocker, c) for c in cpy]
    ind_spy = [spy_task(mocker, i) for i in ind]
    with StageLockContext():
        result = flow.run(**kwargs)
        assert all(result.get(c, as_type=pd.DataFrame)["x"].iloc[0] == 0 for c in cpy)
        out_spy[0].assert_called_once()  # lazy task is always called
        out_spy[1].assert_called_once()  # version=None task is always called
        if (
            mode
            in [
                CacheValidationMode.FORCE_FRESH_INPUT,
                CacheValidationMode.FORCE_CACHE_INVALID,
            ]
            and not disable_cache_function
        ):
            out_spy[2].assert_called(2)
        else:
            out_spy[2].assert_called_once()  # version=AUTO_VERSION task is called once
        if (
            mode
            not in [
                CacheValidationMode.FORCE_FRESH_INPUT,
                CacheValidationMode.FORCE_CACHE_INVALID,
            ]
            and not ignore_task_version
        ):
            out_spy[3].assert_not_called()
        else:
            out_spy[3].assert_called_once()
        if mode not in [CacheValidationMode.FORCE_CACHE_INVALID]:
            if ignore_task_version:
                all(spy.assert_called_once() for spy in cpy_spy)
            elif disable_cache_function:
                if mode == CacheValidationMode.IGNORE_FRESH_INPUT:
                    cpy_spy[0].assert_not_called()  # lazy cache_fn_hash is loaded
                    cpy_spy[1].assert_called_once()  # version=None is always called
                    cpy_spy[2].assert_not_called()  # cache_key is loaded
                    cpy_spy[3].assert_not_called()  # cache_key is loaded
                else:
                    all(spy.assert_called_once() for spy in cpy_spy)
            else:
                cpy_spy[0].assert_not_called()
                cpy_spy[1].assert_called_once()  # version=None is always called
                cpy_spy[2].assert_not_called()  # cache_fn_hash is stable
                cpy_spy[3].assert_not_called()
            if ignore_task_version:
                ind_spy[0].assert_called_once()
            else:
                ind_spy[0].assert_not_called()
            ind_spy[1].assert_called_once()  # lazy is always called
        else:
            all(spy.assert_called_once() for spy in cpy_spy)
            all(spy.assert_called_once() for spy in ind_spy)
    if disable_cache_function:
        assert cache_calls[0] == 0
    else:
        assert cache_calls[0] == 4

    # Changing the cache value should cause it to get called again
    cache_value = 1
    with StageLockContext():
        result = flow.run(**kwargs)
        if mode == CacheValidationMode.IGNORE_FRESH_INPUT and not ignore_task_version:
            assert all(
                result.get(c, as_type=pd.DataFrame)["x"].iloc[0] == res
                for c, res in zip(cpy, [cache_value, cache_value, cache_value, 0])
            )
        else:
            assert all(result.get(c, as_type=pd.DataFrame)["x"].iloc[0] == cache_value for c in cpy)
        out_spy[0].assert_called_once()  # lazy task is always called
        out_spy[1].assert_called_once()  # version=None task is always called
        if disable_cache_function and mode in [
            CacheValidationMode.FORCE_FRESH_INPUT,
            CacheValidationMode.FORCE_CACHE_INVALID,
        ]:
            out_spy[2].assert_called_once()
        else:
            out_spy[2].assert_called(2)  # version=AUTO_VERSION task is called twice
        if mode == CacheValidationMode.IGNORE_FRESH_INPUT and not ignore_task_version:
            out_spy[3].assert_not_called()
        else:
            out_spy[3].assert_called_once()
        if mode not in [CacheValidationMode.FORCE_CACHE_INVALID]:
            all(spy.assert_called_once() for spy in cpy_spy)
            if ignore_task_version:
                ind_spy[0].assert_called_once()
            else:
                ind_spy[0].assert_not_called()
            ind_spy[1].assert_called_once()  # lazy is always called
        else:
            all(spy.assert_called_once() for spy in cpy_spy)
            all(spy.assert_called_once() for spy in ind_spy)


@skip_instances("postgres", "local_table_cache", "snowflake")
@pytest.mark.skipif(dask is None, reason="AUTO_VERSION for pandas requires dask")
@pytest.mark.parametrize("ignore_task_version", [False])
@pytest.mark.parametrize("disable_cache_function", [False])
@pytest.mark.parametrize("mode", ["NORMAL", "IGNORE_FRESH_INPUT", "FORCE_FRESH_INPUT", "FORCE_CACHE_INVALID"])
@pytest.mark.parametrize("imperative", [True])
def test_cache_validation_mode_reduced(ignore_task_version, disable_cache_function, mode, imperative, mocker):
    # Reduce combinatorial space for duckdb to avoid timeout after 10min
    # duckdb is particularly slow for those tests (~10x: 7-15s instead of 1-2s).
    # This reduced combinatorial space is even too much for snowflake. There,
    # we have 1-2 seconds roundtrip time for many individual database requests.
    # And even just reflecting a SQLAlchemy tables issues several of those.
    # Thus this test takes >30min alone for snowflake. In realistic scenarios,
    # the 1-2s per request should not matter too much, though.
    test_cache_validation_mode(ignore_task_version, disable_cache_function, mode, imperative, mocker)


@pytest.mark.parametrize("n", [1, 2, 15])
def test_partial_stage_cache_valid(mocker, n):
    cache_value = 0

    def cache():
        return cache_value

    @materialize(lazy=True, cache=cache)
    def return_cache_table():
        return Table(select_as(cache_value, "x"))

    @materialize(version="1.0")
    def n_tables(n):
        return [pd.DataFrame(dict(x=[i])) for i in range(n)]

    @materialize(lazy=True)
    def n_lazy_tables(n):
        return [Table(select_as(i, "x")) for i in range(n)]

    def get_flow(version):
        @materialize(version=version, input_type=pd.DataFrame)
        def named_copy(table: pd.DataFrame):
            return Table(table, "_table_copy")

        with Flow() as flow:
            with Stage("stage_1"):
                # these tasks are cache valid in all but the first call:
                tbls = n_tables(n)
                lazy_tbls = n_lazy_tables(n)
                # these tables may be cache invalidated by cache_value
                out = return_cache_table()
                cpy = named_copy(out)
        return flow, out, cpy, tbls, lazy_tbls

    flow, out, cpy, tbls, lazy_tbls = get_flow(version="1.0")
    # Initial Call
    with StageLockContext():
        result = flow.run()
        assert result.get(cpy)["x"].iloc[0] == 0

    # Calling flow.run again shouldn't call the task
    out_spy = spy_task(mocker, out)
    cpy_spy = spy_task(mocker, cpy)
    tbls_spy = spy_task(mocker, tbls)
    lazy_tbls_spy = spy_task(mocker, lazy_tbls)
    with StageLockContext():
        result = flow.run()
        assert result.get(cpy)["x"].iloc[0] == 0
        out_spy.assert_called_once()  # lazy task is always called
        cpy_spy.assert_not_called()
        tbls_spy.assert_not_called()
        lazy_tbls_spy.assert_called_once()  # lazy task is always called

    # Changing the cache value should cause it to get called again
    # In this case tbls and lazy_tbls are still cache valid and thus need to be
    # copied over to transaction schema
    cache_value = 1
    with StageLockContext():
        result = flow.run()
        assert result.get(cpy)["x"].iloc[0] == 1
        out_spy.assert_called_once()
        cpy_spy.assert_called_once()
        tbls_spy.assert_not_called()
        lazy_tbls_spy.assert_called_once()  # lazy task is always called


def test_ignore_task_version(mocker):
    cfg = ConfigContext.get().evolve(cache_validation=dict(ignore_task_version=True))

    with Flow() as flow:
        with Stage("stage_1"):
            out = m.noop([1])
            child = m.noop2(out)

    # Initial Call
    with StageLockContext():
        result = flow.run(config=cfg)
        assert result.get(out)[0] == 1
        assert result.get(child)[0] == 1

    # Calling flow.run again should still call the task (disabled caching)
    out_spy = spy_task(mocker, out)
    child_spy = spy_task(mocker, child)
    for _ in range(3):
        with StageLockContext():
            result = flow.run(config=cfg)
            assert result.get(out)[0] == 1
            assert result.get(child)[0] == 1
            out_spy.assert_called_once()
            child_spy.assert_called_once()


@pytest.mark.polars
def test_lazy_table_without_query_string(mocker):
    value = None

    @materialize(lazy=True, nout=5)
    def falsely_lazy_task():
        return (
            Table(pd.DataFrame({"x": [value]}), name="pd_table"),
            Table(pl.DataFrame({"x": [value]}), name="pl_table"),
            Table(pl.DataFrame({"x": [value]}).lazy(), name="pl_lazy_table"),
            select_as(2, "y"),
            3,
        )

    def get_flow():
        with Flow() as flow:
            with Stage("stage_1"):
                pd_tbl, pl_tbl, pl_lazy_tbl, select_tbl, constant = falsely_lazy_task()
                res_pd = m.take_first(pd_tbl, as_int=True)
                res_pl = m.take_first(pl_tbl, as_int=True)
                res_pl_lazy = m.take_first(pl_lazy_tbl, as_int=True)
                res_select = m.noop(select_tbl)
                res_constant = m.noop(constant)
        return flow, res_pd, res_pl, res_pl_lazy, res_select, res_constant

    value = 0
    flow, res_pd, res_pl, res_pl_lazy, res_select, res_constant = get_flow()
    with StageLockContext():
        result = flow.run()
        assert result.get(res_pd) == 0
        assert result.get(res_pl) == 0
        assert result.get(res_pl_lazy) == 0

    value = 1
    flow, res_pd, res_pl, res_pl_lazy, res_select, res_constant = get_flow()
    res_pd_spy = spy_task(mocker, res_pd)
    res_pl_spy = spy_task(mocker, res_pl)
    res_pl_lazy_spy = spy_task(mocker, res_pl_lazy)
    select_spy = spy_task(mocker, res_select)
    constant_spy = spy_task(mocker, res_constant)
    with StageLockContext():
        result = flow.run()
        assert result.get(res_pd) == 1
        assert result.get(res_pl) == 1
        assert result.get(res_pl_lazy) == 1
        # res_pd is downstream of a pd.DataFrame from a lazy task,
        # which should always be cache invalid. Hence, it should always be called.
        res_pd_spy.assert_called_once()

        # res_pd is downstream of a pl.DataFrame from a lazy task,
        # which should always be cache invalid. Hence, it should always be called.
        res_pl_spy.assert_called_once()

        # res_pd is downstream of a pl.LazyFrame from a lazy task,
        # which should always be cache invalid. Hence, it should always be called.
        # To avoid cache-invalidating the LazyFrame, we should use AUTOVERSION.
        res_pl_lazy_spy.assert_called_once()

        # res_select is downstream of an SQL query `select_tbl` (which did not change)
        # from a lazy task, hence select_tbl should be cache valid and the task
        # producing res_select should not be called.
        select_spy.assert_not_called()

        # res_constant is downstream of the constant `constant` (which did not change)
        # from a lazy task, hence res_constant should be cache valid and the task
        # producing res_select should not be called.
        constant_spy.assert_not_called()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from .base import OrchestrationEngine
from .dask import DaskEngine

# don't import prefect engines by default because importing prefect messes with
# initialization of logging library
# from .prefect import PrefectEngine, PrefectOneEngine, PrefectTwoEngine
from .sequential import SequentialEngine

__all__ = [
    "OrchestrationEngine",
    # "PrefectEngine",
    # "PrefectOneEngine",
    # "PrefectTwoEngine",
    "SequentialEngine",
    "DaskEngine",
]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import random
import time
import uuid

import attr
import pytest
import sqlalchemy as sa

from pydiverse.common.util.hashing import stable_hash
from pydiverse.pipedag import (
    Flow,
    GroupNode,
    Result,
    Stage,
    VisualizationStyle,
    materialize,
)
from pydiverse.pipedag.context import FinalTaskState
from pydiverse.pipedag.context.context import (
    ConfigContext,
    StageLockContext,
    TaskContext,
)
from pydiverse.pipedag.core.config import PipedagConfig
from pydiverse.pipedag.core.flow import pydot
from tests.fixtures.instances import (
    ORCHESTRATION_INSTANCES,
    skip_instances,
    with_instances,
)
from tests.util import tasks_library as m
from tests.util.baseline import BaselineStore

pytestmark = [with_instances("postgres", ORCHESTRATION_INSTANCES)]


@pytest.mark.parametrize("ordering_barrier", [True, False])
def test_run_specific_task(ordering_barrier):
    store = ConfigContext.get().store.table_store
    table_name = stable_hash(ordering_barrier, store.engine.url)

    def cache():
        # ensure different position hash for each task
        return uuid.uuid4().hex

    @materialize(cache=cache)
    def task():
        time.sleep(random.randint(0, 1) * 0.1)
        engine = ConfigContext.get().store.table_store.engine
        with engine.connect() as conn:
            with conn.begin():
                conn.execute(sa.text(f'LOCK TABLE "{table_name}" IN ACCESS EXCLUSIVE MODE'))
                conn.execute(sa.text(f'INSERT INTO "{table_name}" VALUES (1)'))
                n = conn.execute(sa.text(f'SELECT COUNT(*) FROM "{table_name}"')).fetchone()[0]
        TaskContext.get().task.logger.info(f"Task result:{n}")
        return n

    # We need to assign unique names to these stages, because we can't reuse the
    # same stage lock context between different runs.
    with Flow() as f:
        with Stage("stage_1"):
            # smoke test
            with GroupNode(ordering_barrier=ordering_barrier):
                _ = m.noop(1)
            _ = m.noop(2)

        with Stage("stage_2"):
            # test barrier ordering
            x1 = task()
            x2 = task()
            with GroupNode(ordering_barrier=ordering_barrier):
                x3 = task()
                x4 = task()
            x5 = task()

        with Stage("stage_3"):
            # smoke test
            _ = m.noop(3)
            with GroupNode(ordering_barrier=ordering_barrier):
                _ = m.noop(4)

        with Stage("stage_4"):
            # smoke test
            with GroupNode(ordering_barrier=ordering_barrier):
                _ = m.noop(5)

    barriers = 0
    if ordering_barrier:
        barriers = 1 + 2 + 2 + 1
    assert len(f.tasks) == 2 + 5 + 2 + 1 + len(f.stages) + barriers

    store.execute(f'DROP TABLE IF EXISTS "{table_name}" ')
    store.execute(f'CREATE TABLE IF NOT EXISTS "{table_name}" (x INT)')

    with StageLockContext():
        res = f.run()
        assert sum(res.get(x) for x in [x1, x2, x3, x4, x5]) == sum(range(1, 6))
        if ordering_barrier:
            assert res.get(x1) < res.get(x3)
            assert res.get(x2) < res.get(x3)
            assert res.get(x1) < res.get(x4)
            assert res.get(x2) < res.get(x4)
            assert res.get(x3) < res.get(x5)
            assert res.get(x4) < res.get(x5)

    with StageLockContext():
        res = f.run(x1, x3)
        assert sum(res.get(x) for x in [x1, x3]) == sum(range(6, 8))
        if ordering_barrier:
            assert res.get(x1) < res.get(x3)

    store.execute(f'DROP TABLE "{table_name}"')


def fake_cache_status_deterministic_for_baseline_tests(result: Result):
    def fix_state(task, state):
        if state in [FinalTaskState.CACHE_VALID, FinalTaskState.COMPLETED]:
            # chose stable pseudo-random
            return (
                FinalTaskState.COMPLETED
                if sum([ord(c) for c in stable_hash(task.id)]) % 2 == 0
                else FinalTaskState.CACHE_VALID
            )

    return attr.evolve(
        result,
        task_states={task: fix_state(task, state) for task, state in result.task_states.items()},
    )


@with_instances("postgres")
@skip_instances(ORCHESTRATION_INSTANCES)
@pytest.mark.parametrize("label", [None, "group"])
@pytest.mark.parametrize(
    "style",
    [
        None,
        VisualizationStyle(hide_box=True),
        VisualizationStyle(hide_box=True, hide_content=True),
        VisualizationStyle(hide_content=True),
        VisualizationStyle(),
    ],
)
@pytest.mark.parametrize("ordering_barrier", [True, False])
@pytest.mark.parametrize("nesting", [False])
@pytest.mark.skipif(pydot.Dot is None, reason="pydot is not installed")
def test_run_specific_task_sequential(label, style, ordering_barrier, nesting):
    num = [0]

    def cache():
        return num[0]

    @materialize(cache=cache)
    def task():
        time.sleep(random.randint(0, 1) * 0.1)
        # random parameter is needed to make sure task is called multiple times
        # despite identical position hash
        num[0] += 1
        return num[0]

    # We need to assign unique names to these stages, because we can't reuse the
    # same stage lock context between different runs.
    with Flow() as f:
        with GroupNode(label, style, ordering_barrier=ordering_barrier):
            with Stage("stage_1"):
                one = m.noop(1)
        with Stage("stage_2"):
            _ = m.noop(one)
            x1 = task()
            x2 = task()
            with GroupNode(label, style, ordering_barrier=ordering_barrier):
                x3 = task()
                x4 = task()
            x5 = task()
        if nesting:
            with GroupNode(label, style, ordering_barrier=ordering_barrier):
                with Stage("stage_3"):
                    with Stage("stage_4"):
                        with GroupNode(label, style, ordering_barrier=ordering_barrier):
                            with GroupNode(label, style, ordering_barrier=ordering_barrier):
                                _ = m.noop(1)
                            _ = m.noop(2)
                        _ = m.noop(3)
                    _ = m.noop(4)
                with Stage("stage_5"):
                    _ = m.noop(5)
            with GroupNode(label, style, ordering_barrier=ordering_barrier):
                with Stage("stage_6"):
                    with Stage("stage_7"):
                        with GroupNode(label, style, ordering_barrier=ordering_barrier):
                            with GroupNode(label, style, ordering_barrier=ordering_barrier):
                                _ = m.noop(6)

    if not nesting:
        barriers = 0
        if ordering_barrier:
            barriers = 2
        assert len(f.tasks) == 2 + 5 + len(f.stages) + barriers

    random.seed(0)  # needed for Baseline comparisons of visualize_url() calls
    cfg = ConfigContext.get().evolve(disable_kroki=False)
    with StageLockContext():
        res = f.run(config=cfg)
        assert res.get(x1) == 1
        assert res.get(x2) == 2
        assert res.get(x3) == 3
        assert res.get(x4) == 4
        assert res.get(x5) == 5
        viz_res = fake_cache_status_deterministic_for_baseline_tests(res)
        assert BaselineStore("flow", label, style, ordering_barrier, nesting) == f.visualize_url(viz_res)

    with StageLockContext():
        res = f.run(x1, x3, config=cfg)
        assert res.get(x1) == 6
        assert res.get(x3) == 7
        viz_res = fake_cache_status_deterministic_for_baseline_tests(res)
        assert BaselineStore("subflow", label, style, ordering_barrier, nesting) == res.subflow.visualize_url(viz_res)


@with_instances("postgres")
@skip_instances(ORCHESTRATION_INSTANCES)
@pytest.mark.parametrize("label", ["group"])
@pytest.mark.parametrize(
    "style",
    [
        VisualizationStyle(),
        VisualizationStyle(hide_label=True),
        VisualizationStyle(hide_content=True, hide_label=True),
        VisualizationStyle(box_color_always="#cc88cc"),
        VisualizationStyle(hide_content=True, box_color_always="#cc88cc"),
        VisualizationStyle(hide_content=True, box_color_all_cache_valid="#cc88cc"),
    ],
)
@pytest.mark.skipif(pydot.Dot is None, reason="pydot is not installed")
def test_run_specific_task_sequential_styles(label, style):
    test_run_specific_task_sequential(label, style, ordering_barrier=False, nesting=True)


@with_instances("postgres")
@skip_instances(ORCHESTRATION_INSTANCES)
@pytest.mark.parametrize("label", [None, "group"])
@pytest.mark.parametrize(
    "style",
    [
        None,
        VisualizationStyle(hide_box=True),
        VisualizationStyle(hide_box=True, hide_content=True),
        VisualizationStyle(hide_content=True),
        VisualizationStyle(hide_content=True, box_color_any_cache_valid="#cc88cc"),
        VisualizationStyle(box_color_always="#cc88cc"),
        VisualizationStyle(),
    ],
)
@pytest.mark.skipif(pydot.Dot is None, reason="pydot is not installed")
def test_run_specific_task_config(label, style):
    group_nodes = dict(
        group=dict(tasks=["task4"]),
        noop=dict(tasks=["noop"]),
        stage=dict(stages=["stage_4"]),
    )
    visualization = dict(default=dict(group_nodes=group_nodes), alternative={})
    if label:
        for group_node in group_nodes.values():
            group_node["label"] = label
    if style:
        visualization["default"]["styles"] = dict(
            test_style={
                k: style.__dict__[k] for k in style.__dataclass_fields__.keys() if style.__dict__[k] is not None
            }
        )
        for group_node in group_nodes.values():
            group_node["style_tag"] = "test_style"

    pipedag_config = PipedagConfig.default
    raw_cfg = pipedag_config.raw_config.copy()
    raw_cfg["instances"]["__any__"]["visualization"] = visualization
    cfg = PipedagConfig(raw_cfg).get().evolve(disable_kroki=False)
    num = [0]

    def cache():
        return num[0]

    @materialize(cache=cache)
    def task():
        time.sleep(random.randint(0, 1) * 0.1)
        # random parameter is needed to make sure task is called multiple times
        # despite identical position hash
        num[0] += 1
        return num[0]

    @materialize(cache=cache, group_node_tag="group")
    def task3():
        return task()

    @materialize(cache=cache)
    def task4():
        return task()

    @materialize
    def noop2(x):
        return m.noop(x)

    # We need to assign unique names to these stages, because we can't reuse the
    # same stage lock context between different runs.
    with Flow() as f:
        with Stage("stage_1"):
            x1 = task()
            x2 = task()
            x3 = task3()
            x4 = task4()
            x5 = task()
        with Stage("stage_2", group_node_tag="stage"):
            with Stage("stage_3"):
                _ = m.noop(6)
                with GroupNode("style_tag", style_tag="test_style"):
                    _ = noop2(1)
                with GroupNode("no_style_tag"):
                    _ = noop2(2)
        with Stage("stage_4"):
            _ = noop2(7)

    barriers = 0
    assert len(f.tasks) == 4 + 5 + len(f.stages) + barriers

    random.seed(0)  # needed for Baseline comparisons of visualize_url() calls
    with StageLockContext():
        res = f.run(config=cfg)
        assert res.get(x1) == 1
        assert res.get(x2) == 2
        assert res.get(x3) == 3
        assert res.get(x4) == 4
        assert res.get(x5) == 5
        viz_res = fake_cache_status_deterministic_for_baseline_tests(res)
        assert BaselineStore("config_flow", label, style) == f.visualize_url(viz_res)
        assert BaselineStore("config2_flow", label, style) == f.visualize_url(viz_res, "alternative")

    with StageLockContext():
        res = f.run(x1, x3, config=cfg)
        assert res.get(x1) == 6
        assert res.get(x3) == 7
        viz_res = fake_cache_status_deterministic_for_baseline_tests(res)
        assert BaselineStore("config_subflow", label, style) == res.subflow.visualize_url(viz_res)
        assert BaselineStore("config2_subflow", label, style) == res.subflow.visualize_url(viz_res, "alternative")



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from . import cache
from .dict import DictTableStore
from .sql import SQLTableStore

__all__ = [
    "cache",
    "DictTableStore",
    "SQLTableStore",
]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import uuid

import pandas as pd
import sqlalchemy as sa
import structlog

from pydiverse.pipedag import Flow, Stage, materialize
from pydiverse.pipedag.context import ConfigContext
from tests.fixtures.instances import with_instances


@with_instances("postgres", "postgres_unlogged")
def test_postgres_unlogged():
    @materialize(version="1.0.0")
    def dataframe(manual_invalidate):
        _ = manual_invalidate
        return pd.DataFrame({"x": [1]})

    @materialize(lazy=True)
    def sql_table(manual_invalidate):
        _ = manual_invalidate
        return sa.select(sa.literal(1).label("x"))

    @materialize(input_type=sa.Table)
    def get_relpersistence(table: sa.sql.expression.Alias):
        return sa.text(
            """
            SELECT relpersistence
            FROM pg_class
            LEFT JOIN pg_namespace ON pg_class.relnamespace = pg_namespace.oid
            WHERE nspname = :schema
              AND relname = :name
            """
        ).bindparams(
            schema=str(table.original.schema),
            name=str(table.original.name),
        )

    @materialize(input_type=pd.DataFrame)
    def assert_relpersistence(df: pd.DataFrame):
        relpersistence = (
            "u" if ConfigContext.get().store.table_store.materialization_details["__any__"].unlogged else "p"
        )
        assert df["relpersistence"][0] == relpersistence

    def get_flow(manual_invalidate, partial_invalidate):
        with Flow() as f:
            with Stage("stage"):
                df = dataframe(manual_invalidate)
                tbl = sql_table(manual_invalidate)
                # just to prevent 100% cache validity
                _ = sql_table(partial_invalidate)
            with Stage("check"):
                rp_df = get_relpersistence(df)
                rp_tbl = get_relpersistence(tbl)
                assert_relpersistence(rp_df)
                assert_relpersistence(rp_tbl)
        return f

    manual_invalidate = str(uuid.uuid4())
    partial_invalidate = str(uuid.uuid4())

    logger = structlog.get_logger("test_postgres_unlogged")
    logger.info("1st run")
    f = get_flow(manual_invalidate, partial_invalidate)
    f.run()

    logger.info("2nd run with 100% cache valid stage")
    f.run()

    logger.info("3rd run with partial cache invalid stage")
    partial_invalidate = str(uuid.uuid4())
    f = get_flow(manual_invalidate, partial_invalidate)
    f.run()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import networkx as nx
import pytest

from pydiverse.pipedag import Flow, Stage
from pydiverse.pipedag.core import UnboundTask
from pydiverse.pipedag.core.flow import pydot
from pydiverse.pipedag.errors import DuplicateNameError, FlowError, StageError


def t(name: str, **kwargs):
    return UnboundTask((lambda *a: a), name=f"task-{name}", **kwargs)


def t_kw(name: str, **kwargs):
    return UnboundTask((lambda **kw: kw), name=f"task-kw-{name}", **kwargs)


def validate_dependencies(flow: Flow):
    # Tasks outside a stage can only run if all
    # tasks inside the stage have finished.

    g = flow.graph
    expl_g = flow.build_graph()

    assert nx.is_directed_acyclic_graph(g)
    assert nx.is_directed_acyclic_graph(expl_g)

    stages = flow.stages.values()
    tasks = flow.tasks

    # Check flow.graph is correct
    for task in tasks:
        assert task in g
        assert task in expl_g

        parents = {edge[0] for edge in g.in_edges(task)}
        assert set(task.input_tasks.values()) == parents

    # Check inputs computed before task
    for task in tasks:
        for input_task in task.input_tasks.values():
            assert nx.shortest_path(expl_g, input_task, task)

    # Check each task in stage happens before commit
    for task in tasks:
        assert nx.shortest_path(expl_g, task, task.stage.commit_task)

    # Check commit task dependencies
    for child in tasks:
        for parent, _ in g.in_edges(child):  # type: Task
            if child.stage == parent.stage:
                continue

            if child.stage.is_inner(parent.stage):
                continue

            assert nx.shortest_path(expl_g, parent, parent.stage.commit_task)
            assert nx.shortest_path(expl_g, parent.stage.commit_task, child)

    # Ensure that nested stages get committed before their parents
    for stage in stages:
        if stage.outer_stage is None:
            continue
        assert nx.shortest_path(expl_g, stage.commit_task, stage.outer_stage.commit_task)


class TestDAGConstruction:
    """
    Test that the DAG gets constructed properly, all metadata is set
    properly and that all task dependencies are correct.
    """

    def test_one_task(self):
        with Flow("f") as f:
            with Stage("s") as s:
                t0 = t("0")(0)

        assert f.stages == {"s": s}
        assert f.tasks == [t0, s.commit_task]

        assert s.tasks == [t0]
        assert s.outer_stage is None

        assert t0.upstream_stages == []

        validate_dependencies(f)

    def test_two_independent_tasks(self):
        with Flow("f") as f:
            with Stage("s") as s:
                t0 = t("0")(0)
                t1 = t("1")(1)

        assert f.stages == {"s": s}
        assert f.tasks == [t0, t1, s.commit_task]

        assert s.tasks == [t0, t1]
        assert s.outer_stage is None

        assert t0.upstream_stages == []
        assert t1.upstream_stages == []

        validate_dependencies(f)

    def test_two_connected_tasks(self):
        with Flow("f") as f:
            with Stage("s") as s:
                t0 = t("0")(0)
                t1 = t("1")(t0)

        assert f.stages == {"s": s}
        assert f.tasks == [t0, t1, s.commit_task]

        assert s.tasks == [t0, t1]
        assert s.outer_stage is None

        assert t0.upstream_stages == []
        assert t1.upstream_stages == [s]

        validate_dependencies(f)

    def test_two_stages(self):
        with Flow("f") as f:
            with Stage("stage_0") as s0:
                t00 = t("00")(0)
                t01 = t("01")(t00)

            with Stage("stage_1") as s1:
                t10 = t("10")(t00)
                t11 = t("11")(t10, t00)
                t12 = t("12")(t00)

        assert f.stages == {"stage_0": s0, "stage_1": s1}
        assert f.tasks == [t00, t01, s0.commit_task, t10, t11, t12, s1.commit_task]

        assert s0.tasks == [t00, t01]
        assert s0.outer_stage is None

        assert s1.tasks == [t10, t11, t12]
        assert s1.outer_stage is None

        assert t00.upstream_stages == []
        assert t01.upstream_stages == [s0]
        assert t10.upstream_stages == [s0]
        assert t11.upstream_stages == [s1, s0]
        assert t12.upstream_stages == [s0]

        validate_dependencies(f)

    def test_nested(self):
        with Flow("f") as f:
            with Stage("stage_0") as s0:
                t00 = t("00")(0)
                t01 = t("01")(t00)

            with Stage("stage_1") as s1:
                t10 = t("10")(t00)
                t11 = t("11")(t10, t00)
                t12 = t("12")(t00)

                with Stage("stage_2") as s2:
                    t20 = t("20")(t11)

            with Stage("stage_3") as s3:
                t30 = t("30")(t12, t20)
                t31 = t("31")(t00, t10, t20, t30)

        assert f.stages == {"stage_0": s0, "stage_1": s1, "stage_2": s2, "stage_3": s3}
        assert f.tasks == [
            t00,
            t01,
            s0.commit_task,
            t10,
            t11,
            t12,
            t20,
            s2.commit_task,
            s1.commit_task,
            t30,
            t31,
            s3.commit_task,
        ]

        assert s0.tasks == [t00, t01]
        assert s0.outer_stage is None

        assert s1.tasks == [t10, t11, t12]
        assert s1.outer_stage is None

        assert s2.tasks == [t20]
        assert s2.outer_stage is s1
        assert s2.is_inner(s1)

        assert s3.tasks == [t30, t31]
        assert s3.outer_stage is None

        assert t00.upstream_stages == []
        assert t01.upstream_stages == [s0]
        assert t10.upstream_stages == [s0]
        assert t11.upstream_stages == [s1, s0]
        assert t12.upstream_stages == [s0]
        assert t20.upstream_stages == [s1]
        assert t30.upstream_stages == [s1, s2]
        assert t31.upstream_stages == [s0, s1, s2, s3]

        validate_dependencies(f)

    def test_ids(self):
        with Flow("f"):
            with Stage("stage_0") as s0:
                t00 = t("00")(0)
                t01 = t("01")(t00)

            with Stage("stage_1") as s1:
                t10 = t("10")(t00)
                t11 = t("11")(t10, t00)
                t12 = t("12")(t00)

                with Stage("stage_2") as s2:
                    t20 = t("20")(t11)

            with Stage("stage_3") as s3:
                t30 = t("30")(t12, t20)
                t31 = t("31")(t00, t10, t20, t30)

        assert [stage.id for stage in [s0, s1, s2, s3]] == [0, 1, 2, 3]
        assert [
            task.id
            for task in [
                t00,
                t01,
                s0.commit_task,
                t10,
                t11,
                t12,
                t20,
                s2.commit_task,
                s1.commit_task,
                t30,
                t31,
                s3.commit_task,
            ]
        ] == list(range(12))

    def test_subflow(self):
        with Flow("f") as f:
            with Stage("stage_0"):
                t00 = t("00")(0)
                t01 = t("01")(t00)
                t02 = t("02")(t01)

            with Stage("stage_1") as s1:
                t10 = t("10")(t02)
                with Stage("stage_2") as s2:
                    t20 = t("20")(t10)
                t11 = t("11")(t20)
                t12 = t("12")(t10, t20, t11)

        # Entire Flow as Subflow
        sf = f.get_subflow()
        assert not sf.is_tasks_subflow
        assert set(sf.get_tasks()) == set(f.tasks)

        # Stage Subflow
        sf = f.get_subflow(s2)
        assert not sf.is_tasks_subflow
        assert set(sf.get_tasks()) == {t20, s2.commit_task}
        assert set(sf.get_parent_tasks(t20)) == set()

        # Stage Subflow
        sf = f.get_subflow(s1)
        assert not sf.is_tasks_subflow
        assert set(sf.get_tasks()) == {
            t10,
            t11,
            t12,
            t20,
            s1.commit_task,
            s2.commit_task,
        }
        assert set(sf.get_parent_tasks(t10)) == set()
        assert set(sf.get_parent_tasks(t11)) == {t20, s2.commit_task}

        # Task Subflow
        sf = f.get_subflow(t10, t12)
        assert sf.is_tasks_subflow
        assert set(sf.get_tasks()) == {t10, t12}
        assert set(sf.get_parent_tasks(t10)) == set()
        assert set(sf.get_parent_tasks(t12)) == {t10}


class TestDAGConstructionExceptions:
    def test_duplicate_stage_name(self):
        with Flow("flow 1"):
            with Stage("stage"):
                # Nested
                with pytest.raises(DuplicateNameError):
                    with Stage("stage"):
                        ...

            # Consecutive
            with pytest.raises(DuplicateNameError):
                with Stage("stage"):
                    ...

            # Different capitalization
            with pytest.raises(DuplicateNameError):
                with Stage("Stage"):
                    ...

        # Should be able to reuse name in different flow
        with Flow("flow 2"):
            with Stage("stage"):
                ...

    def test_reuse_stage(self):
        with Flow("flow 1"):
            with Stage("stage") as s:
                # Nested
                with pytest.raises(StageError):
                    with s:
                        ...

            # Consecutive
            with pytest.raises(StageError):
                with s:
                    ...

        with Flow("flow 2"):
            # Different flow
            with pytest.raises(StageError):
                with s:
                    ...

    def test_stage_outside_flow(self):
        with pytest.raises(StageError):
            with Stage("stage"):
                ...

    def test_task_outside_stage(self):
        with Flow("flow"):
            with pytest.raises(StageError):
                t("task")()

    def test_task_in_wrong_flow(self):
        with Flow("flow 1"):
            with Stage("stage"):
                task = t("task")()

        with Flow("flow 2"):
            with Stage("stage"):
                with pytest.raises(FlowError):
                    # Can't use task from different flow as argument
                    t("bad task")(task)


class TestPositionHash:
    def test_single_task(self):
        with Flow():
            with Stage("stage_0"):
                x = t("0")(0)
                y = t("0")(0)

                z0 = t("0")(1)  # Different input
                z1 = t("1")(0)  # Different name

            with Stage("stage_1"):
                z2 = t("0")(0)  # Different stage

        assert x.position_hash == y.position_hash

        assert x.position_hash != z0.position_hash
        assert x.position_hash != z1.position_hash
        assert x.position_hash != z2.position_hash

    def test_multiple_tasks(self):
        with Flow():
            with Stage("stage_0"):
                x0 = t("0")(0)
                y0 = t("0")(0)

                z0_0 = t("1")(0)  # Wrong task name

            with Stage("stage_1"):
                x1 = t("1")(2, x0)
                y1 = t("1")(2, y0)

                z1_0 = t("1")(2, z0_0)  # z0_0 has different position hash
                z1_1 = t("1")(1, x1)  # Different input: 1 != 2

            with Stage("stage_2"):
                x2 = t_kw("2")(a=1, b=2, c=[x1, y1])
                y2 = t_kw("2")(b=2, a=1, c=[y1, x1])

                z2_0 = t_kw("2")(A=1, b=2, c=[x1, y1])  # Different kwarg name
                z2_1 = t_kw("2")(a=1, b=2, c=[z1_0, z1_1])  # Inputs different pos hash

        assert x0.position_hash == y0.position_hash
        assert x0.position_hash != z0_0.position_hash

        assert x1.position_hash == y1.position_hash
        assert x1.position_hash != z1_0.position_hash
        assert x1.position_hash != z1_1.position_hash

        assert x2.position_hash == y2.position_hash
        assert x2.position_hash != z2_0.position_hash
        assert x2.position_hash != z2_1.position_hash

    def test_get_item(self):
        with Flow():
            with Stage("stage_0"):
                inputs = t_kw("inputs")(a=1, b=2)

                x0 = t("0")(inputs["a"])
                y0 = t("0")(inputs["a"])
                z0 = t("0")(inputs["b"])

            with Stage("stage_1"):
                inputs = t("inputs")(1, 2)

                x1 = t("0")(inputs[0])
                y1 = t("0")(inputs[0])
                z1 = t("0")(inputs[1])

        assert x0.position_hash == y0.position_hash
        assert x0.position_hash != z0.position_hash

        assert x1.position_hash == y1.position_hash
        assert x1.position_hash != z1.position_hash

        assert inputs["1"].position_hash != inputs[1].position_hash
        assert inputs[1].position_hash != inputs[1.0].position_hash
        assert inputs[1][1].position_hash != inputs[1][0].position_hash
        assert inputs[1][1].position_hash != inputs[0][1].position_hash


class TestFlow:
    def test_get_stage(self):
        with Flow() as f:
            with Stage("s0") as s0:
                ...

            with Stage("s1") as s1:
                ...

        assert f.get_stage("s0") == f["s0"] == s0
        assert f.get_stage("s1") == f["s1"] == s1

        with pytest.raises(KeyError):
            f.get_stage("s2")


class TestStage:
    def test_get_task(self):
        with Flow():
            with Stage("s0") as s0:
                t00_0 = t("00")(0)
                t00_1 = t("00")(1)

            with Stage("s1") as s1:
                t10 = t("10")(t00_0)
                t11 = t("11")(t00_1)
                t00_s1 = t("00")(0)

        assert s0.get_task(t00_0.name, 0) == s0[t00_0.name, 0] == t00_0
        assert s0.get_task(t00_1.name, 1) == s0[t00_0.name, 1] == t00_1

        assert s1.get_task(t10.name) == s1[t10.name] == t10
        assert s1.get_task(t10.name, 0) == s1[t10.name, 0] == t10
        assert s1.get_task(t11.name) == t11
        assert s1.get_task(t11.name, 0) == t11
        assert s1.get_task(t00_s1.name) == t00_s1
        assert s1.get_task(t00_s1.name, 0) == t00_s1

        with pytest.raises(LookupError):
            # Task doesn't exist
            s0.get_task("foo")
        with pytest.raises(ValueError):
            # Missing index
            s0.get_task(t00_0.name)
        with pytest.raises(IndexError):
            # Out of bounds
            s0.get_task(t00_0.name, 2)


def test_task_nout():
    with Flow("flow"):
        with Stage("stage"):
            _ = t("task")()
            _ = t("task", nout=1)()

            _, _ = t("task", nout=2)()
            _, _, _ = t("task", nout=3)()
            _, _, *_ = t("task", nout=10)()

            with pytest.raises(ValueError):
                _, _ = t("task")()

            with pytest.raises(ValueError):
                t("task", nout=0)
            with pytest.raises(ValueError):
                t("task", nout=1.5)


def test_task_getitem():
    with Flow("flow"):
        with Stage("stage"):
            task = t("task")()

    assert task.resolve_value((1, 2)) == (1, 2)
    assert task[0].resolve_value((1, 2)) == 1
    assert task[1].resolve_value((1, 2)) == 2
    assert task[1][0].resolve_value(((1, 2), (3, 4))) == 3
    assert task["x"][1:3].resolve_value({"x": [1, 2, 3, 4]}) == [2, 3]


def test_task_outside_flow():
    task = t("task")
    assert task(1, 2) == (1, 2)
    assert task("foo", "bar") == ("foo", "bar")

    task = t_kw("task")
    assert task(a=1, b=2) == {"a": 1, "b": 2}
    assert task(x="foo", y="bar") == {"x": "foo", "y": "bar"}


def test_flow_visualize_url():
    from pydiverse.pipedag.core import PipedagConfig

    with Flow("flow") as f:
        with Stage("stage"):
            _ = t("task")()

    # Kroki is disabled by default
    with PipedagConfig.default.get().evolve():
        if pydot.Dot is not None:
            visualization_url = f.visualize_url()
            assert visualization_url.startswith("<disable_kroki=True>/graphviz/")
        else:
            with pytest.raises(RuntimeError, match="please install pydot"):
                f.visualize_url()

    # Use kroki.io as default url
    with PipedagConfig.default.get().evolve(disable_kroki=False):
        if pydot.Dot is not None:
            visualization_url = f.visualize_url()
            assert visualization_url.startswith("https://kroki.io/graphviz/")
        else:
            with pytest.raises(RuntimeError, match="please install pydot"):
                f.visualize_url()

    # Check that overriding works
    with PipedagConfig.default.get().evolve(disable_kroki=False, kroki_url="THIS_IS_A_TEST_URL"):
        if pydot.Dot is not None:
            visualization_url = f.visualize_url()
            assert visualization_url.startswith("THIS_IS_A_TEST_URL/graphviz/")
        else:
            with pytest.raises(RuntimeError, match="please install pydot"):
                f.visualize_url()
